"""Defines common filters."""

from __future__ import annotations

import os
from functools import lru_cache

from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import (
    Condition,
    emacs_insert_mode,
    emacs_mode,
    to_filter,
    vi_insert_mode,
    vi_mode,
    vi_replace_mode,
)

from euporie.core.key_binding.micro_state import MicroInputMode


@Condition
@lru_cache
def have_black() -> "bool":
    """Determine if black is available."""
    try:
        import black.const  # noqa F401
    except ModuleNotFoundError:
        return False
    else:
        return True


@Condition
@lru_cache
def have_isort() -> "bool":
    """Determine if isort is available."""
    try:
        import isort  # noqa F401
    except ModuleNotFoundError:
        return False
    else:
        return True


@Condition
@lru_cache
def have_ssort() -> "bool":
    """Determine if ssort is available."""
    try:
        import ssort  # noqa F401
    except ModuleNotFoundError:
        return False
    else:
        return True


# Determine if we have at least one formatter
have_formatter = have_black | have_isort | have_ssort

# Determine if euporie is running inside tmux.
in_tmux = to_filter(os.environ.get("TMUX") is not None)


@Condition
def cursor_in_leading_ws() -> "bool":
    """Determine if the cursor of the current buffer is in leading whitespace."""
    from prompt_toolkit.application.current import get_app

    before = get_app().current_buffer.document.current_line_before_cursor
    return (not before) or before.isspace()


@Condition
def has_suggestion() -> "bool":
    """Determine if the current buffer can display a suggestion."""
    from prompt_toolkit.application.current import get_app

    app = get_app()
    return (
        app.current_buffer.suggestion is not None
        and len(app.current_buffer.suggestion.text) > 0
        and app.current_buffer.document.is_cursor_at_the_end_of_line
    )


@Condition
def has_dialog() -> "bool":
    """Determine if a dialog is being displayed."""
    from prompt_toolkit.layout.containers import ConditionalContainer

    from euporie.notebook.app import get_app

    app = get_app()
    for dialog in app.dialogs.values():
        if isinstance(dialog.content, ConditionalContainer):
            if dialog.content.filter():
                return True
    return False


@Condition
def has_menus() -> "bool":
    """Determine if a menu is being displayed."""
    from prompt_toolkit.layout.containers import ConditionalContainer

    from euporie.notebook.app import get_app

    app = get_app()
    for menu in app.menus.values():
        if isinstance(menu.content, ConditionalContainer):
            if menu.content.filter():
                return True
    return False


@Condition
def tab_has_focus() -> "bool":
    """Determine if there is a currently focused tab."""
    from euporie.core.app import get_app

    return get_app().tab is not None


@Condition
def pager_has_focus() -> "bool":
    """Determine if there is a currently focused notebook."""
    from euporie.core.app import get_app

    app = get_app()
    pager = app.pager
    if pager is not None:
        return app.layout.has_focus(pager)
    return False


@Condition
def display_has_focus() -> "bool":
    """Determine if there is a currently focused cell."""
    from euporie.core.app import get_app
    from euporie.core.widgets.display import DisplayControl

    return isinstance(get_app().layout.current_control, DisplayControl)


@Condition
def buffer_is_empty() -> "bool":
    """Determine if the current buffer contains nothing."""
    from euporie.core.app import get_app

    return not get_app().current_buffer.text


@Condition
def buffer_is_code() -> "bool":
    """Determine if the current buffer contains code."""
    from euporie.core.app import get_app

    return get_app().current_buffer.name == "code"


@Condition
def buffer_is_markdown() -> "bool":
    """Determine if the current buffer contains markdown."""
    from euporie.core.app import get_app

    return get_app().current_buffer.name == "code"


@Condition
def micro_mode() -> "bool":
    """When the micro key-bindings are active."""
    from euporie.core.app import get_app

    return get_app().editing_mode == EditingMode.MICRO  # type: ignore


@Condition
def micro_replace_mode() -> "bool":
    """Determine if the editor is in overwrite mode."""
    from euporie.core.app import get_app

    app = get_app()
    return app.micro_state.input_mode == MicroInputMode.REPLACE


@Condition
def micro_insert_mode() -> "bool":
    """Determine if the editor is in insert mode."""
    from euporie.core.app import get_app

    app = get_app()
    return app.micro_state.input_mode == MicroInputMode.INSERT


@Condition
def micro_recording_macro() -> "bool":
    """Determine if a micro macro is being recorded."""
    from euporie.core.app import get_app

    return get_app().micro_state.current_recording is not None


@Condition
def is_returnable() -> "bool":
    """Determine if the current buffer has an accept handler."""
    from euporie.core.app import get_app

    return get_app().current_buffer.is_returnable


@Condition
def cursor_at_start_of_line() -> "bool":
    """Determine if the cursor is at the start of a line."""
    from euporie.core.app import get_app

    return get_app().current_buffer.document.cursor_position_col == 0


@Condition
def cursor_on_first_line() -> "bool":
    """Determine if the cursor is on the first line of a buffer."""
    from euporie.core.app import get_app

    return get_app().current_buffer.document.on_first_line


@Condition
def cursor_on_last_line() -> "bool":
    """Determine if the cursor is on the last line of a buffer."""
    from euporie.core.app import get_app

    return get_app().current_buffer.document.on_last_line


"""Determine if any binding style is in insert mode."""
insert_mode = (
    (vi_mode & vi_insert_mode)
    | (emacs_mode & emacs_insert_mode)
    | (micro_mode & micro_insert_mode)
)

"""Determine if any binding style is in replace mode."""
replace_mode = micro_replace_mode | vi_replace_mode


@Condition
def is_searching() -> "bool":
    """Determine if the app is in search mode."""
    from euporie.core.app import get_app

    app = get_app()
    return (
        app.search_bar is not None and app.search_bar.control in app.layout.search_links
    )


@Condition
def at_end_of_buffer() -> "bool":
    """Determine if the cursor is at the end of the current buffer."""
    from prompt_toolkit.application.current import get_app

    buffer = get_app().current_buffer
    return buffer.cursor_position == len(buffer.text)


@Condition
def kernel_is_python() -> "bool":
    """Determine if the current notebook has a python kernel."""
    from euporie.core.app import get_app
    from euporie.core.tabs.base import KernelTab

    kernel_tab = get_app().tab
    if isinstance(kernel_tab, KernelTab):
        return kernel_tab.language == "python"
    return False


@Condition
def multiple_cells_selected() -> "bool":
    """Determine if there is more than one selected cell."""
    from euporie.core.app import get_app
    from euporie.core.tabs.notebook import BaseNotebook

    nb = get_app().tab
    if isinstance(nb, BaseNotebook):
        return len(nb.selected_indices) > 1
    return False


@Condition
def kernel_tab_has_focus() -> "bool":
    """Determine if there is a focused kernel tab."""
    from euporie.core.app import get_app
    from euporie.core.tabs.base import KernelTab

    return isinstance(get_app().tab, KernelTab)



"""Contains functions which convert data to PIL format."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from euporie.core.convert.base import register
from euporie.core.convert.utils import have_modules

if TYPE_CHECKING:
    from typing import Optional

    from PIL import Image
    from upath import UPath

log = logging.getLogger(__name__)


def set_background(image: "Image", bg_color: "Optional[str]" = None) -> "bytes":
    """Removes the alpha channel from an image and set the background colour."""
    from PIL import Image

    if image.mode in ("RGBA", "LA") or (
        image.mode == "P" and "transparency" in image.info
    ):
        alpha = image.convert("RGBA").getchannel("A")
        bg = Image.new("RGBA", image.size, bg_color)
        bg.paste(image, mask=alpha)
        image = bg
    image = image.convert("P", palette=Image.ADAPTIVE, colors=16).convert(
        "RGB", palette=Image.ADAPTIVE, colors=16
    )
    return image


@register(
    from_=("png", "jpeg", "gif"),
    to="pil",
    filter_=have_modules("PIL"),
)
def png_to_pil_py(
    data: "bytes",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "Image":
    """Convert PNG to a pillow image using :py:mod:`PIL`."""
    import io

    from PIL import Image

    try:
        image = Image.open(io.BytesIO(data))
    except IOError:
        log.error("Could not load image.")
    else:
        return image



"""Defines custom inputs and outputs, and related methods."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING

from prompt_toolkit.input import vt100_parser
from prompt_toolkit.input.base import DummyInput, _dummy_context_manager
from prompt_toolkit.output.vt100 import Vt100_Output as PtkVt100_Output

if TYPE_CHECKING:
    from typing import Any, Callable, ContextManager, Union

    from prompt_toolkit.keys import Keys

_response_prefix_re = re.compile(
    r"""^\x1b(
        \][^\\\x07]*  # Operating System Commands
        |
        _[^\\]*  # Application Program Command
        |
        \[\?[\d;]*  # Primary device attribute responses
    )\Z""",
    re.VERBOSE,
)


class _IsPrefixOfLongerMatchCache(vt100_parser._IsPrefixOfLongerMatchCache):
    def __missing__(self, prefix: "str") -> "bool":
        """Check if the response might match an OSC or APC code, or DA response."""
        result = super().__missing__(prefix)
        if not result:
            if _response_prefix_re.match(prefix):
                result = True
                self[prefix] = result
        return result


# Monkey patch the prefix cache
vt100_parser._IS_PREFIX_OF_LONGER_MATCH_CACHE = _IsPrefixOfLongerMatchCache()


class Vt100Parser(vt100_parser.Vt100Parser):
    """A Vt100Parser which checks input against additional key patterns."""

    def __init__(self, *args: "Any", **kwargs: "Any") -> "None":
        """Create a new VT100 parser."""
        super().__init__(*args, **kwargs)
        self.queries: "dict[Keys, re.Pattern]" = {}

    def _get_match(self, prefix: "str") -> "Union[None, Keys, tuple[Keys, ...]]":
        """Checks for additional key matches first."""
        for key, pattern in self.queries.items():
            if pattern.match(prefix):
                return key

        return super()._get_match(prefix)


class IgnoredInput(DummyInput):
    """An input which ignores input but does not immediately close the app."""

    def attach(
        self, input_ready_callback: "Callable[[], None]"
    ) -> "ContextManager[None]":
        """Do not call the callback, so the input is never closed."""
        return _dummy_context_manager()


class Vt100_Output(PtkVt100_Output):
    """A Vt100 output which enables SGR pixel mouse positioning."""

    def enable_mouse_support(self) -> "None":
        """Additionally enable SGR-pixel mouse positioning."""
        super().enable_mouse_support()
        self.write_raw("\x1b[?1016h")

    def disable_mouse_support(self) -> "None":
        """Additionally disable SGR-pixel mouse positioning."""
        super().disable_mouse_support()
        self.write_raw("\x1b[?1016l")



"""Run euporie as a multi-client SSH server."""

from __future__ import annotations

import logging
from asyncio import get_event_loop
from importlib.metadata import entry_points
from typing import TYPE_CHECKING

try:
    import asyncssh
except ModuleNotFoundError as err:
    raise ModuleNotFoundError(
        "Euporie hub requires the the `asyncssh` python package\n"
        'Install euporie hub with: "pip install euporie[hub]"'
    ) from err

from prompt_toolkit.contrib.ssh import PromptToolkitSSHSession
from upath import UPath

from euporie.core.app import BaseApp
from euporie.core.config import add_setting
from euporie.core.log import setup_logs

if TYPE_CHECKING:
    from typing import Type

log = logging.getLogger(__name__)


class EuporieSSHServer(asyncssh.SSHServer):  # type: ignore
    """Launch euporie hub, which serves a euporie app over SSH.

    Launches euporie hub, a multi-client SSH server running euporie, which
    allows multiple users to connect and run instances of a euporie app.

    """

    def __init__(self, app_cls: "Type[BaseApp]") -> None:
        """Sets the interaction function for the SSH session."""
        self.app_cls = app_cls

    def begin_auth(self, username: str) -> bool:
        """Perform authentication in the SSH server."""
        if self.app_cls.config.no_auth:
            # No authentication.
            return False
        return super().begin_auth(username)

    def session_requested(self) -> "PromptToolkitSSHSession":
        """Return an SSH session."""
        # Not sure why mypy gives an error here
        return PromptToolkitSSHSession(
            self.app_cls.interact, enable_cpr=True
        )  # type: ignore [call-arg]


class HubApp(BaseApp):
    """Hub App.

    An app which runs as a multi-user SSH server.

    This app never actually gets run, but is used to run another app in an SSH server.
    """

    @classmethod
    def launch(cls) -> "None":
        """Launch the HubApp SSH server."""
        # Configure some setting defaults
        cls.config.settings["log_file"].value = "-"
        cls.config.settings[
            "log_config"
        ].value = '{"loggers": {"asyncssh": {"handlers":["stdout"], "level": "DEBUG"}}}'

        # Load the app's configuration
        cls.config.load(cls)
        # Configure the logs
        setup_logs(cls.config)

        if cls.config.no_auth:
            log.warning(
                "This server has been configured without SSH authentication, "
                "meaning anyone can connect"
            )

        # Import the hubbed app
        if entry_point := {
            entry.name: entry for entry in entry_points()["euporie.apps"]
        }.get(cls.config.app):
            app_cls = entry_point.load()
        else:
            raise ValueError("Application `%s` not found", cls.config.app)

        # Run the HubApp in an SSH server
        loop = get_event_loop()
        loop.run_until_complete(
            asyncssh.create_server(
                lambda: EuporieSSHServer(app_cls=app_cls),
                host=cls.config.host,
                port=cls.config.port,
                server_host_keys=cls.config.host_keys,
                authorized_client_keys=cls.config.client_keys,
            )
        )
        log.info("Running euporie hub on port %s", cls.config.port)
        loop.run_forever()

    # ################################### Settings ####################################

    add_setting(
        name="app",
        flags=["--app"],
        default="notebook",
        type_=str,
        choices=["notebook", "console"],
        help_="App to run under euporie hub",
        description="""
            Determine which euporie app should be launched under euporie hub.
        """,
    )

    add_setting(
        name="host",
        flags=["--host"],
        type_=str,
        help_="The host address to bind to",
        default="",
        description="""
            This determines the host address the euporie hub SSH server will bind to.
        """,
    )

    add_setting(
        name="port",
        flags=["--port"],
        type_=int,
        help_="The port for the ssh server to use",
        default=8022,
        schema_={
            "minimum": 1,
            "maximum": 65535,
        },
        description="""
            This determines which port euporie will listen on for connections to
            euporie hub.
        """,
    )

    add_setting(
        name="host_keys",
        flags=["--host-keys"],
        nargs="*",
        type_=UPath,
        help_="Host keys to use for the SSH server",
        default=["/etc/ssh/ssh_host_ecdsa_key"],
        schema={
            "type": "array",
            "items": {
                "description": "SSH host key file path",
                "type": "string",
            },
        },
        description="""
            One or more SSH host key files to use for the euporie hub SSH server.
        """,
    )

    add_setting(
        name="client_keys",
        flags=["--client-keys"],
        nargs="*",
        type_=UPath,
        help_="Client public keys authorized to connect",
        default=["~/.ssh/authorized_keys"],
        schema={
            "type": "array",
            "items": {
                "description": "Path to file containing authorized public keys",
                "type": "string",
            },
        },
        description="""
            One or more OpenSSH-style :file:`authorized_keys` files, containing
            public keys for authorized clients.
        """,
    )

    add_setting(
        name="no_auth",
        flags=["--no-auth"],
        type_=bool,
        help_="Allow unauthenticated access to euporie hub",
        default=False,
        description="""
            When set, users will be able to access euporie hub without authentication.

            .. warning::

               This option is dangerous, as arbitrary code can be executed through
               euporie apps.
        """,
    )



"""Contains a HTML to formatted text parser."""

from __future__ import annotations

import logging
import re
from ast import literal_eval
from collections import ChainMap, defaultdict
from functools import cached_property, lru_cache, partial
from html.parser import HTMLParser
from math import ceil
from random import randint
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.application.current import get_app_session
from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.formatted_text.base import StyleAndTextTuples
from prompt_toolkit.formatted_text.utils import fragment_list_width, split_lines

from euporie.core.border import (
    BorderLineStyle,
    Double,
    Invisible,
    Padding,
    Rounded,
    Thick,
    Thin,
)
from euporie.core.convert.base import convert, get_format
from euporie.core.convert.utils import data_pixel_size, pixels_to_cell_size
from euporie.core.formatted_text.table import Table
from euporie.core.formatted_text.utils import (
    FormattedTextAlign,
    add_border,
    align,
    apply_style,
    indent,
    last_line_length,
    lex,
    max_line_width,
    paste,
    strip,
    truncate,
    wrap,
)
from euporie.core.terminal import tmuxify
from euporie.core.url import load_url

if TYPE_CHECKING:
    from typing import Any, Generator, Hashable, Optional, Union

log = logging.getLogger(__name__)


class Position(NamedTuple):
    """A description of a floating element's location."""

    top: "Optional[int]" = None
    right: "Optional[int]" = None
    bottom: "Optional[int]" = None
    left: "Optional[int]" = None


class Direction(NamedTuple):
    """A description of a direction."""

    x: "bool" = False
    y: "bool" = False


# Prefer 6-digit hex-colors over 3-digit ones
_COLOR_RE = re.compile("#([a-fA-F0-9]{6}|[a-fA-F0-9]{3})")

# List of elements which might not have a close tag
_VOID_ELEMENTS = (
    "area",
    "base",
    "br",
    "col",
    "embed",
    "hr",
    "img",
    "input",
    "link",
    "meta",
    "source",
    "track",
    "wbr",
)


_BORDER_WIDTHS = {
    "0": BorderLineStyle(Invisible, Invisible, Invisible, Invisible),
    "1": BorderLineStyle(Thin, Thin, Thin, Thin),
    "2": BorderLineStyle(Thick, Thick, Thick, Thick),
}


def _inline_default() -> "dict[str, Any]":
    return {"inline": True, "block": False, "margin": Padding(0, 0, 0, 0)}


# The default theme to apply to various elements
_ELEMENT_BASE_THEMES: "defaultdict[str, dict[str, Any]]" = defaultdict(
    dict,
    {
        # Default theme for a tag
        "default": {
            "padding": Padding(None, None, None, None),
            "margin": Padding(0, 0, 0, 0),
            "align": FormattedTextAlign.LEFT,
            "border": None,
            "border_collapse": True,
            "language": None,
            "style": "",
            "invisible": False,
            "skip": False,
            "block": True,
            "float": False,
            "zindex": 0,
            "position": Position(None, None, None, None),
            "inline": False,
            # Styles
            "style_classes": [],
        },
        # "Special" tags (not real HTML tags, just used for rendering)
        "text": {  # Use for rendering text
            "inline": True,
            "block": False,
            "margin": Padding(0, 0, 0, 0),
        },
        # Metadata elements which should be hidden
        "head": {"skip": True},
        "base": {"skip": True},
        "command": {"skip": True},
        "link": {"skip": True},
        "meta": {"skip": True},
        "noscript": {"skip": True, "block": False},
        "script": {"skip": True, "block": False},
        "style": {"skip": True},
        "title": {"skip": True},
        # Inline tags
        "a": {"inline": True, "block": False},
        "abbr": {"inline": True, "block": False},
        "acronym": {"inline": True, "block": False},
        "audio": {"inline": True, "block": False},
        "b": {"inline": True, "block": False},
        "bdi": {"inline": True, "block": False},
        "bdo": {"inline": True, "block": False},
        "big": {"inline": True, "block": False},
        "br": {"inline": True, "block": False},
        "button": {"inline": True, "block": False},
        "canvas": {"inline": True, "block": False},
        "cite": {"inline": True, "block": False},
        "code": {"inline": True, "block": False},
        "data": {"inline": True, "block": False},
        "datalist": {"inline": True, "block": False},
        "del": {"inline": True, "block": False},
        "dfn": {"inline": True, "block": False},
        "em": {"inline": True, "block": False},
        "embed": {"inline": True, "block": False},
        "i": {"inline": True, "block": False},
        "iframe": {"inline": True, "block": False},
        "img": {"inline": True, "block": False},
        "input": {"inline": True, "block": False},
        "ins": {"inline": True, "block": False},
        "kbd": {"inline": True, "block": False},
        "label": {"inline": True, "block": False},
        "map": {"inline": True, "block": False},
        "mark": {"inline": True, "block": False},
        "meter": {"inline": True, "block": False},
        "object": {"inline": True, "block": False},
        "output": {"inline": True, "block": False},
        "picture": {"inline": True, "block": False},
        "progress": {"inline": True, "block": False},
        "q": {"inline": True, "block": False},
        "ruby": {"inline": True, "block": False},
        "s": {"inline": True, "block": False},
        "samp": {"inline": True, "block": False},
        "select": {"inline": True, "block": False},
        "slot": {"inline": True, "block": False},
        "small": {"inline": True, "block": False},
        "span": {"inline": True, "block": False},
        "strong": {"inline": True, "block": False},
        "sub": {"inline": True, "block": False},
        "sup": {"inline": True, "block": False},
        "svg": {"inline": True, "block": False},
        "template": {"inline": True, "block": False},
        "textarea": {"inline": True, "block": False},
        "time": {"inline": True, "block": False},
        "u": {"inline": True, "block": False},
        "tt": {"inline": True, "block": False},
        "var": {"inline": True, "block": False},
        "video": {"inline": True, "block": False},
        "wbr": {"inline": True, "block": False},
        # Custom inline tags
        "math": {
            "inline": True,
            "block": False,
            "align": FormattedTextAlign.CENTER,
            "margin": Padding(1, 0, 1, 0),
        },
        # Alignment
        "center": {"align": FormattedTextAlign.CENTER},
        # Table elements
        "table": {"padding": Padding(0, 1, 0, 1)},
        "td": {"block": True, "inline": True, "align": FormattedTextAlign.LEFT},
        "th": {
            "block": True,
            "inline": True,
            "style": "bold",
            "align": FormattedTextAlign.LEFT,
        },
        # Forms & related elements
        "option": {"skip": True},
        # Custom default styles
        "blockquote": {"margin": Padding(1, 0, 1, 0)},
        "hr": {"margin": Padding(1, 0, 1, 0)},
        "p": {"margin": Padding(1, 0, 1, 0)},
        "pre": {"margin": Padding(1, 0, 1, 0)},
        "details": {"margin": Padding(1, 0, 1, 0)},
        "summary": {"margin": Padding(0, 0, 1, 0)},
        "caption": {"align": FormattedTextAlign.CENTER},
        # Lists
        "ol": {"list-style": "decimal", "block": True},
        "ul": {"list-style": "disc", "block": True},
    },
)

_ELEMENT_INSETS = {
    "blockquote": 2,
    "details": 3,
    "summary": 3,
}


def try_eval(value: "str", default: "Any" = None) -> "Any":
    """Attempt to cast a string to a python type."""
    parsed_value = default or value
    try:
        parsed_value = literal_eval(value)
    except (
        ValueError,
        TypeError,
        SyntaxError,
        MemoryError,
        RecursionError,
    ):
        pass
    return parsed_value


def get_integer(value: "str") -> "str":
    """Extract the first integer from a string."""
    for word in value.split(" "):
        for c in word.partition(".")[0].split():
            if c.isdigit():
                return c.lstrip("0")
    return ""


def get_color(value: "str") -> "str":
    """Extract a hex color from a string."""
    color = ""
    if match := re.search(_COLOR_RE, value):
        hexes = match.group(1)
        if len(hexes) == 3:
            hexes = "".join(2 * s for s in hexes)
        color = f"#{hexes}"
    elif value.startswith("rgb"):
        # Ignore alpha for now - TODO
        color = "#" + "".join(
            (hex(int(x))[2:] for x in value.strip("rgba()").split(",")[:3])
        )
    else:
        from euporie.core.reference import NAMED_COLORS

        color = NAMED_COLORS.get(value, "")
    return color


class PageElement:
    """Represents an HTML element."""

    def __init__(
        self,
        name: "str",
        parent: "Optional[PageElement]",
        attrs: "list[tuple[str, Optional[str]]]",
        text: "str" = "",
    ) -> "None":
        """Create a new page element."""
        self.name = name
        self.parent = parent
        self.text = text
        self.attrs: "dict[str, Any]" = {k: v for k, v in attrs if v is not None}
        self.contents: "list[PageElement]" = []
        self.closed = False
        # self.attrs["class"] = self.attrs.get("class", "").split()

    def find_all(self, tag: "str", recursive: "bool" = False) -> "list[PageElement]":
        """Find all child elements of a given tag type."""
        return [element for element in self.contents if element.name == tag]

    @property
    def descendents(self) -> "Generator[PageElement, None, None]":
        """Yields all descendent elements."""
        for child in self.contents:
            yield child
            yield from child.descendents

    @cached_property
    def parents(self) -> "list[PageElement]":
        """Yields all descendent elements."""
        parents = []
        parent = self.parent
        while parent is not None:
            parents.append(parent)
            parent = parent.parent
        return parents[::-1]

    def _outer_html(self, d: "int" = 0, attrs: "bool" = True) -> "str":
        dd = " " * d
        s = ""
        if self.name != "text":
            s += f"{dd}<{self.name}"
            if attrs:
                for key, value in self.attrs.items():
                    if key == "class":
                        value = " ".join(value)
                    if value:
                        s += f' {key}="{value}"'
            s += ">"
            if self.contents:
                for child in self.contents:
                    s += f"\n{dd} {child._outer_html(d+1, attrs=attrs)}"
                s += f"\n{dd}{dd}"
            s += f"</{self.name}>"
        else:
            s += f"{dd}{self.text}"
        return s

    def __repr__(self, d: "int" = 0) -> "str":
        return self._outer_html(attrs=False)


class CustomHTMLParser(HTMLParser):
    """An HTML parser."""

    def __init__(self) -> "None":
        """Create a new parser instance."""
        super().__init__()
        self.curr = self.soup = PageElement(name="html", parent=None, attrs=[])

    def parse(self, markup: "str") -> "PageElement":
        """Parse HTML markup."""
        self.curr = self.soup = PageElement(name="html", parent=None, attrs=[])
        self.feed(markup)
        return self.soup

    def handle_starttag(
        self, tag: "str", attrs: "list[tuple[str, Optional[str]]]"
    ) -> "None":
        """Open a new element."""
        self.autoclose()
        element = PageElement(name=tag, parent=self.curr, attrs=attrs)
        self.curr.contents.append(element)
        self.curr = element

    def autoclose(self) -> "None":
        """Automatically close void elements."""
        if not self.curr.closed and self.curr.name in _VOID_ELEMENTS:
            self.curr.closed = True
            if self.curr.parent:
                self.curr = self.curr.parent

    def handle_data(self, data: "str") -> "None":
        """Create data (text) elements."""
        self.autoclose()
        self.curr.contents.append(
            PageElement(name="text", parent=self.curr, text=data, attrs=[])
        )

    def handle_endtag(self, tag: "str") -> "None":
        """Handle end tags: close the currently opened element."""
        if tag != self.curr.name:
            self.autoclose()
        self.curr.closed = True
        if self.curr.parent:
            self.curr = self.curr.parent


parser = CustomHTMLParser()


def css_dimension(value: "str", vertical: "bool" = False) -> "Optional[int|float]":
    """Convert CSS dimensions to terminal cell sizes."""
    # Get digits
    digits = ""
    i = 0
    try:
        while (c := value[i]) in "0123456789.":
            digits += c
            i += 1
    except IndexError:
        pass
    # Ensure we only have one decimal place
    digits = ".".join(digits.split(".")[:2])
    # Convert to Float
    try:
        number = float(digits)
    except ValueError:
        return None

    if number == 0:
        return 0

    # Get units
    units = ""
    i = -1
    try:
        while (c := value[i]) not in "0123456789.":
            units += c
            i -= 1
    except IndexError:
        pass
    units = units[::-1]

    # Calculate size based on units
    if vertical:
        if units == "px":
            cols, aspect = pixels_to_cell_size(px=number, py=number)
            # Add 0.5 to round to nearest whole number
            return int((cols * aspect) + 0.5)
        elif units == "%":
            return number / 100
        else:
            return int(number)

    else:
        if units == "px":
            cols, _ = pixels_to_cell_size(px=number, py=1)
            return cols
        elif units == "%":
            return number / 100
        else:
            # Round down for vertical space
            return int(number) * 2


def parse_css_content(content: "str") -> "dict[str, Any]":
    """Convert CSS declarations into the internals style representation."""
    output: "dict[str, Any]" = {}

    for declaration in content.split(";"):
        name, _, value = declaration.partition(":")
        name = name.strip()
        # Ignore "!important" tags for now - TODO
        value = value.replace("!important", "").strip()

        if name == "color":
            if color := get_color(value):
                output.setdefault("style_fg", "")
                output["style_fg"] = color

        elif name in ("background", "background-color"):
            if color := get_color(value):
                output.setdefault("style_bg", "")
                output["style_bg"] = color

        elif name == "font-weight":
            if value in ("bold", "bolder") or (
                (digits := get_integer(value)) and int(digits) > 700
            ):
                output.setdefault("style_attrs", [])
                output["style_attrs"].append("bold")

        elif name == "text-decoration":
            output.setdefault("style_attrs", [])
            if value == "underline":
                output["style_attrs"].append("underline")
            if value == "none":
                output["style_attrs"].append("nounderline")

        elif name == "text-align":
            value = value.lower()
            if value == "left":
                output["align"] = FormattedTextAlign.LEFT
            elif value == "center":
                output["align"] = FormattedTextAlign.CENTER
            elif value == "right":
                output["align"] = FormattedTextAlign.RIGHT

        elif name == "border-width":
            if digits := get_integer(value):
                output["border"] = _BORDER_WIDTHS.get(
                    digits,
                    BorderLineStyle(Thick, Thick, Thick, Thick),
                )

        elif name == "visibility":
            output["hidden"] = value == "hidden"

        elif name == "display":
            if value == "flex":
                output["flex"] = True
            if value == "block":
                output["block"] = True
            if value == "inline-block":
                output["block"] = True
                output["inline"] = True
            elif value == "none":
                output["skip"] = True

        elif name == "flex-direction":
            output["flex-direction"] = value

        elif name == "position":
            # if value == "absolute":
            # output["float"] = True
            if value == "fixed":
                output["float"] = True

        elif name in ("width", "max-width"):
            # Get width in terminal colmuns
            if width := css_dimension(value, vertical=False):
                output["width"] = width

        elif name == "height":
            # Get height in terminal rows
            if height := css_dimension(value, vertical=True):
                output["height"] = height

        elif name in ("top", "right", "bottom", "left"):
            current = output.get("position", Position(None, None, None, None))._asdict()
            if (
                dim := css_dimension(value, vertical=name in ("top", "bottom"))
            ) is not None:
                output["position"] = Position(**{**current, **{name: dim}})

        elif name == "padding":
            top = right = bottom = left = 0
            values = value.split()
            if len(values) == 1:
                top = bottom = css_dimension(values[0], vertical=True) or 0
                right = left = css_dimension(values[0], vertical=False) or 0
            if len(values) == 2:
                top = bottom = css_dimension(values[0], vertical=True) or 0
                left = right = css_dimension(values[1], vertical=False) or 0
            elif len(values) == 3:
                top = css_dimension(values[0], vertical=True) or 0
                right = left = css_dimension(values[1], vertical=False) or 0
                bottom = css_dimension(values[2], vertical=True) or 0
            elif len(values) == 4:
                top = css_dimension(values[0], vertical=True) or 0
                right = css_dimension(values[1], vertical=False) or 0
                bottom = css_dimension(values[2], vertical=True) or 0
                left = css_dimension(values[3], vertical=False) or 0
            output["padding"] = Padding(top, right, bottom, left)

        elif name.startswith("padding-"):
            prop = name[8:]
            if prop in ("top", "right", "bottom", "left"):
                current = output.get("padding", Padding(0, 0, 0, 0))._asdict()
                if dim := css_dimension(value, vertical=name in ("top", "bottom")):
                    output["padding"] = Padding(**{**current, **{prop: dim}})

        elif name == "margin":
            top = right = bottom = left = 0
            values = value.split()
            if len(values) == 1:
                top = bottom = right = left = values[0]
            elif len(values) == 2:
                top = bottom = values[0]
                left = right = values[1]
            elif len(values) == 3:
                top = values[0]
                right = left = values[1]
                bottom = values[2]
            elif len(values) == 4:
                top, right, bottom, left = values

            # Align to center if margin-x is auto
            if left == "auto" and right == "auto":
                output["align"] = FormattedTextAlign.CENTER
            elif left == "auto":
                output["align"] = FormattedTextAlign.RIGHT
            elif right == "auto":
                output["align"] = FormattedTextAlign.LEFT

            output["margin"] = Padding(
                css_dimension(top, vertical=True) or 0,
                css_dimension(right, vertical=False) or 0,
                css_dimension(bottom, vertical=True) or 0,
                css_dimension(left, vertical=False) or 0,
            )

        elif name.startswith("margin-"):
            prop = name[7:]
            if prop in ("top", "right", "bottom", "left"):
                current = output.get("padding", Padding(0, 0, 0, 0))._asdict()
                if dim := css_dimension(value, vertical=prop in ("top", "bottom")):
                    output["margin"] = Padding(**{**current, **{prop: dim}})

        elif name == "overflow" and value == "hidden":
            output["truncate"] = Direction(x=True, y=True)
        elif name == "overflow-x" and value == "hidden":
            current = output.get("truncate", Direction(x=False, y=False))._asdict()
            output["truncate"] = Direction(**{**current, **{"x": True}})
        elif name == "overflow-y" and value == "hidden":
            current = output.get("truncate", Direction(x=False, y=False))._asdict()
            output["truncate"] = Direction(**{**current, **{"y": True}})

        elif name == "opacity" and value == "0":
            output["invisible"] = True

        elif name == "list-style-type" or name == "list-style":
            if "none" in value.split():
                value = ""
            output["list-style"] = value

    return output


class HTML:
    """A HTML formatted text renderer.

    Accepts a HTML string and renders it at a given width.
    """

    def _render_list(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Render lists, adding item numbers to child <li> elements."""
        theme = self.element_theme(element, parent_theme)
        items = [li for li in element.contents if li.name == "li"]
        for i, item in enumerate(items, start=1):
            item.attrs["data-list-index"] = i
            item.attrs["data-list-length"] = len(items)
        # Render contents as normal
        return self.render_contents(
            element.contents,
            theme,
            available_width,
            available_height,
            left,
            preformatted,
        )

    render_ul = _render_list
    render_ol = _render_list

    def __init__(
        self,
        markup: "str",
        base: "Optional[Union[UPath, str]]" = None,
        width: "Optional[int]" = None,
        height: "Optional[int]" = None,
        strip_trailing_lines: "bool" = True,
        pad: "bool" = True,
    ) -> None:
        """Initialize the markdown formatter.

        Args:
            markup: The markdown text to render
            base: The base url for the HTML document
            width: The width in characters available for rendering. If :py:const:`None`
                the terminal width will be used
            height: The width in characters available for rendering. If :py:const:`None`
                the terminal height will be used
            strip_trailing_lines: If :py:const:`True`, empty lines at the end of the
                rendered output will be removed
            pad: When :py:const:`True`, the output is padded to fill the width

        """
        self.markup = markup
        self.base = base
        self.strip_trailing_lines = strip_trailing_lines
        self.pad = pad
        self.width: "Optional[int]" = None
        self.height: "Optional[int]" = None

        self.css: "dict[str, dict[str, Any]]" = {
            ".dataframe": {
                "border": BorderLineStyle(Invisible, Invisible, Invisible, Invisible),
                "border_collapse": True,
                "style_classes": ["dataframe"],
            },
            ".block": {"block": True},
        }

        self.element_theme_cache: "SimpleCache[Hashable, dict[str, Any]]" = (
            SimpleCache()
        )

        # Parse the markup
        self.soup = parser.parse(markup.strip())

        # Parse the styles
        self.css.update(self.parse_styles(self.soup))

        # Render the markup
        self.render(width, height)

        if strip_trailing_lines:
            self.formatted_text = strip(
                self.formatted_text,
                left=False,
                char="\n",
            )

    def render(self, width: "int", height: "int") -> "StyleAndTextTuples":
        """Render the current markup at a given size."""
        if not width or not height:
            size = get_app_session().output.get_size()
        self.width = width or size.columns
        self.height = height or size.rows

        self.formatted_text = self.render_element(
            self.soup,
            parent_theme={},
            available_width=self.width,
            available_height=self.height,
        )

    def parse_styles(self, soup: "PageElement") -> "dict[str, dict[str, str]]":
        """Collect all CSS styles from style tags."""
        rules: "dict[str, dict[str, str]]" = {}
        for child in soup.descendents:
            css_str = ""

            # In case of a <link> style, load the url
            if (
                child.name == "link"
                and child.attrs.get("rel") == "stylesheet"
                and (href := child.attrs.get("href"))
            ):
                if css_bytes := load_url(href, self.base):
                    css_str = css_bytes.decode()

            # In case of a <style> tab, load first child's text
            elif child.name == "style":
                if child.contents:
                    css_str = child.contents[0].text
            else:
                continue
            css_str = css_str.strip().replace("\n", "")
            # Remove comments
            css_str = re.sub(r"\/\*[^\*]+\*\/", "", css_str)
            # Allow plain '@media screen' queries
            css_str = re.sub(
                r"""
                @media\s+screen\s*{\s*
                  (.+? {
                    (?:[^:}]+\s*:\s*[^;}]+\s*;\s*)*
                    (?:[^:}]+\s*:\s*[^;}]+\s*;?\s*)?
                  })
                  \s*
                }
                """,
                "\\1",
                css_str,
                0,
                flags=re.DOTALL | re.VERBOSE,
            )
            # Remove other media queries for now - TODO
            css_str = re.sub(
                r"""
                @media.+?{\s*
                  (.+? {
                    (?:[^:}]+\s*:\s*[^;}]+\s*;\s*)*
                    (?:[^:}]+\s*:\s*[^;}]+\s*;?\s*)?
                  })
                  \s*
                }
                """,
                "",
                css_str,
                0,
                flags=re.DOTALL | re.VERBOSE,
            )
            if css_str:
                css_str = css_str.replace("\n", "").strip()
                if css_str:
                    for rule in css_str.rstrip("}").split("}"):
                        selectors, _, content = rule.partition("{")
                        # TODO - more CSS matching complex rules
                        for selector in map(str.strip, selectors.split(",")):
                            content = content.strip().rstrip(";")
                            rule_content = parse_css_content(content)
                            if rule_content:
                                if selector in rules:
                                    rules[selector].update(rule_content)
                                else:
                                    rules[selector] = rule_content

        return rules

    def element_theme(
        self, element: "PageElement", parent_theme: "dict[str, Any]"
    ) -> "dict":
        """Get an element's theme from the cache, or calculate it."""
        return self.element_theme_cache.get(
            # TODO - check this key is unique enough
            (
                element.name,
                element.attrs.get("id"),
                tuple(element.attrs.get("class", [])),
                element.attrs.get("style"),
                tuple(parent_theme.get("style_classes", [])),
            ),
            partial(self.calc_element_theme, element, parent_theme),
        )

    def calc_element_theme(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
    ) -> "dict[str, Any]":
        """Compute the theme of an element."""
        # Add extra attributes
        extras: "dict[str, Any]" = {}
        styles: "dict[str, Any]" = {}

        # -> border
        if border_attr := element.attrs.get("border"):
            extras["border"] = _BORDER_WIDTHS.get(
                border_attr,
                BorderLineStyle(Thick, Thick, Thick, Thick),
            )
        # -> bgcolor
        if bg := element.attrs.get("bgcolor"):
            color = get_color(bg)
            if color:
                extras["style_bg"] = color
        # -> width
        if (value := element.attrs.get("width")) is not None:
            extras["width"] = css_dimension(value)
        # -> cellpadding
        # TODO

        # Check for language highlighting classes
        for class_name in element.attrs.get("class", "").split():
            if class_name.startswith("language-"):
                extras["language"] = class_name[9:]
                break
            elif class_name.startswith("highlight-"):
                extras["language"] = class_name[10:]
                break

        # for class_name in element.attrs.get("class", []):
        # if (rule := self.css.get(f"{element.name}.{class_name}")) is not None:
        # class_rules.append(rule)
        # if (rule := self.css.get(f".{class_name}")) is not None:
        # class_rules.append(rule)
        # self.css.get(element.name, {}),

        @lru_cache
        def _match_selector(selector: "str", element: "PageElement") -> "bool":
            matched = True

            while selector and matched:

                # Universal selector
                if selector == "*":
                    break

                # Type selectors
                if selector[0] not in ".#[":
                    if selector.startswith(name := element.name):
                        selector = selector[len(name) :]
                    else:
                        matched = False

                # ID selectors
                if selector.startswith("#"):
                    if "id" in element.attrs:
                        id_ = element.attrs["id"]
                        if selector[1:].startswith(id_):
                            selector = selector[1 + len(id_) :]
                        else:
                            matched = False
                    else:
                        matched = False

                # Class selectors
                if selector and selector.startswith("."):
                    for class_name in sorted(
                        element.attrs.get("class", "").split(), key=len, reverse=True
                    ):
                        # IF CLASS NAME SUBSTRING
                        if selector[1:].startswith(class_name):
                            selector = selector = selector[1 + len(class_name) :]
                            break
                    else:
                        matched = False

                # Attribute selectors
                if selector.startswith("["):
                    test = selector[1 : selector.index("]")]
                    if (op := "*=") in test:
                        attr, _, value = test.partition(op)
                        value = value.strip("'\"")
                        matched = value in element.attrs.get(attr, "")
                    elif (op := "$=") in test:
                        attr, _, value = test.partition(op)
                        value = value.strip("'\"")
                        matched = element.attrs.get(attr, "").endswith(value)
                    elif (op := "^=") in test:
                        attr, _, value = test.partition(op)
                        value = value.strip("'\"")
                        matched = element.attrs.get(attr, "").startswith(value)
                    elif (op := "|=") in test:
                        attr, _, value = test.partition(op)
                        value = value.strip("'\"")
                        matched = element.attrs.get(attr) in (value, f"{value}-")
                    elif (op := "~=") in test:
                        attr, _, value = test.partition(op)
                        value = value.strip("'\"")
                        matched = value in element.attrs.get(attr, "").split()
                    elif (op := "=") in test:
                        attr, _, value = test.partition(op)
                        value = value.strip("'\"")
                        matched = element.attrs.get(attr) == value
                    else:
                        matched = test in element.attrs
                    selector = selector[2 + len(test) :]

                if not matched:
                    break

            return matched

        css_rules = {}
        for selectors, rule in reversed(self.css.items()):
            split_selectors = selectors.split()
            parent_selectors = split_selectors[:-1]
            element_selector = split_selectors[-1]
            for selector in parent_selectors:
                for parent in element.parents:
                    if _match_selector(selector, parent):
                        break
                else:
                    break
            else:
                if _match_selector(element_selector, element):
                    css_rules[selectors] = rule

        # Chain themes from various levels
        theme = ChainMap(
            # Computed combined styles
            styles,
            # Tag styles
            parse_css_content(element.attrs.get("style", "")),
            # Class rules
            *css_rules.values(),
            # ID rule
            # self.css.get(f"#{element_id}", {}),
            # Tag rules
            # self.css.get(element.name, {}),
            # Tag attributes
            extras,
            # Element base style
            _ELEMENT_BASE_THEMES.get(element.name, _ELEMENT_BASE_THEMES["default"]),
            # Parent theme
            dict(parent_theme),
            # Add an element class
            {"style_classes": [element.name]},
            # Default element style
            _ELEMENT_BASE_THEMES["default"],
        )

        # Concatenate styles
        styles["style_classes"] = sum(
            (mapping.get("style_classes", []) for mapping in theme.maps), start=[]
        )
        styles["style_attrs"] = sum(
            (mapping.get("style_attrs", []) for mapping in theme.maps), start=[]
        )
        style_classes_str = "class:" + ",".join(styles["style_classes"])
        style_attrs_str = " ".join(styles["style_attrs"])
        styles["style"] = f"{style_classes_str} {style_attrs_str}"
        if fg := theme.get("style_fg"):
            styles["style"] += f" fg:{fg}"
        if bg := theme.get("style_bg"):
            styles["style"] += f" bg:{bg}"

        # Flatten chained dict
        return theme
        return dict(theme)

    def render_contents(
        self,
        contents: "list[PageElement]",
        parent_theme: "dict[str, Any]",
        available_width: "int" = 80,
        available_height: "int" = 999999999,
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Render a list of parsed markdown elements.

        Args:
            contents: The list of parsed elements to render
            parent_theme: The theme of the element's parent element
            available_width: The width available for rendering the elements
            available_height: The height available for rendering the elements
            left: The position on the current line at which to render the output - used
                to indent subsequent lines when rendering inline blocks like images
            preformatted: If True, whitespace will not be stripped from the element's
                text

        Returns:
            Formatted text

        """
        ft: "StyleAndTextTuples" = []

        def _draw_y_margin(
            ft: "StyleAndTextTuples", element: "PageElement", index: "int"
        ) -> "StyleAndTextTuples":
            """Draw the vertical margins for an element."""
            theme = self.element_theme(element, parent_theme)
            margin = theme["margin"][index]
            # Find how much of this margin has already been drawn
            i = -1
            for frag in reversed(ft):
                for c in reversed(frag[1]):
                    if c == "\n":
                        i += 1
                        continue
                    break
                else:
                    continue
                break

            # Draw remaining margin lines
            if n_new_lines := max(0, margin - i):
                return [
                    (
                        parent_theme["style"] if parent_theme else "",
                        "\n" * n_new_lines,
                    )
                ]
            return []

        outputs = []
        floats = {}

        for i, element in enumerate(contents):

            # Convert tags with "math" class to <math> tag
            if "math" in element.attrs.get("class", []):
                element.name = "math"

            theme = self.element_theme(element, parent_theme)

            # Do not render tags with "skip" set in theme
            if theme["skip"]:
                continue

            # Is the current element a block element?
            block = theme["block"]

            # Set preformatted flag for <pre> tags
            _preformatted = preformatted
            if element.name == "pre":
                _preformatted = True

            # Render block element margins. We want to ensure block elements always
            # start on a new line, and that margins collapse.
            # Do not draw a margin if this is the first element of the render list.
            if block and ft:
                ft.extend(_draw_y_margin(ft, element, 0))

            # If there is a special method for rendering the block, use it; otherwise
            # use the generic `render_element` function
            render_func = getattr(self, f"render_{element.name}", self.render_element)

            # Recalculate the left offset if we have rendered output
            if ft:
                left = last_line_length(ft)

            # Render the element
            rendering = render_func(
                element,
                parent_theme=parent_theme,
                available_width=available_width,
                available_height=available_height,
                left=left,
                preformatted=_preformatted,
            )

            if block:
                # Draw block element bottom margin, ensuring block elements end on a new
                # line, and that margins collapse
                if parent_theme["block"] and any(
                    element.name != "text"
                    or (element.name == "text" and element.text.strip())
                    for element in contents[i + 1 :]
                ):
                    rendering.extend(_draw_y_margin(rendering, element, 2))

            if theme["float"]:
                floats[(theme["zindex"], theme["position"])] = rendering
            else:
                ft.extend(rendering)

        # Draw flex elements
        # if parent_theme.get("flex") and parent_theme.get("flex-direction") == "column":
        # table = Table(border=Invisible, border_collapse=True)
        # row = table.new_row()
        # for output in outputs:
        # row.new_cell(output)
        # ft = table.render(available_width)
        #
        # else:
        ft = sum(outputs, start=ft)

        # Draw floats
        for (_, position), float_ft in sorted(floats.items()):
            row = col = 0
            if (top := position.top) is not None:
                row = top
            elif (bottom := position.bottom) is not None:
                row = (
                    sum(1 for _ in split_lines(ft))
                    - sum(1 for _ in split_lines(float_ft))
                    - bottom
                )
            if (left := position.left) is not None:
                col = left
            elif (right := position.right) is not None:
                row = max_line_width(ft) - max_line_width(float_ft) - right

            ft = paste(ft, float_ft, row, col)

        return ft

    def render_element(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Render a list of parsed markdown elements representing a block element.

        Args:
            element: The list of parsed elements to render
            parent_theme: The theme of the element's parent element
            available_width: The width available for rendering the elements
            available_height: The height available for rendering the elements
            left: The position on the current line at which to render the output - used
                to indent subsequent lines when rendering inline blocks like images
            preformatted: When True, whitespace in the the element's text is not
                collapsed

        Returns:
            Formatted text

        """
        theme = self.element_theme(element, parent_theme)

        inner_width = available_width - _ELEMENT_INSETS.get(element.name, 0)
        inner_height = available_height

        if (theme_height := theme.get("height")) is not None:
            if isinstance(theme_height, float) and 0 < theme_height <= 1:
                inner_height = int(inner_height * theme_height + 0.5)
            else:
                inner_height = min(inner_height, theme_height)

        if (theme_width := theme.get("width")) is not None:
            if isinstance(theme_width, float) and 0 < theme_width <= 1:
                inner_width = int(inner_width * theme_width + 0.5)
            else:
                inner_width = min(inner_width, theme_width)

        if margin := theme.get("margin"):
            for margin_size in (margin.left, margin.right):
                if isinstance(margin_size, float) and 0 < margin_size <= 1:
                    inner_width = int(inner_width * (1 - margin_size))
                else:
                    inner_width -= margin_size or 0

        if padding := theme.get("padding"):
            for padding_size in (padding.left, padding.right):
                if isinstance(padding_size, float) and 0 < padding_size <= 1:
                    inner_width = int(inner_width * (1 - padding_size))
                else:
                    inner_width -= padding_size or 0

        # Render the contents
        ft = self.render_contents(
            element.contents,
            parent_theme=theme,
            available_width=inner_width,
            available_height=inner_height,
            left=left,
            preformatted=preformatted,
        )

        # If an element should not overflow it's width / height, truncate it
        if trunc := theme.get("truncate"):
            if trunc.x:
                ft = truncate(ft, available_width, placeholder="")
            if trunc.y:
                new_ft = []
                for i, line in enumerate(split_lines(ft)):
                    if i <= inner_height:
                        new_ft += line
                        new_ft += [("", "\n")]
                ft = new_ft[:-1]

        # Apply tag formatting
        if format_func := getattr(self, f"format_{element.name}", None):
            ft = format_func(
                ft,
                inner_width,
                inner_height,
                left,
                element,
                theme,
            )

        # Lex the text
        if language := theme["language"]:
            ft = strip(ft, left=False, right=True, char="\n")
            ft = lex(ft, lexer_name=language)

        if theme["block"] and not preformatted:

            # Align the output
            if theme["align"] != FormattedTextAlign.LEFT:
                ft = align(
                    theme["align"], ft, width=available_width, style=theme["style"]
                )

            # Add left margin
            if margin and (margin_left := margin.left):
                if isinstance(margin_left, float) and 0 < margin.left <= 1:
                    margin_left = int(available_width * margin_left)
                ft = indent(ft, margin=" " * margin_left, style=parent_theme["style"])

            # Add left padding
            if padding and (padding_left := padding.left):
                if isinstance(padding_left, float) and 0 < padding.left <= 1:
                    padding_left = int(available_width * padding_left)
                ft = indent(ft, margin=" " * padding_left, style=theme["style"])

            # Fill space around block elements so they fill the width
            if self.pad and not theme["inline"]:

                # Remove one trailing newline if there is one
                for i in range(len(ft) - 1, -1, -1):
                    frag = ft[i]
                    if not frag[1]:
                        continue
                    if frag[1] == "\n":
                        del ft[i]
                    elif frag[1].endswith("\n"):
                        ft[i] = (frag[0], frag[1][:-1])
                    break

                # Format the remainder of each line
                filled_output = []
                for line in split_lines(ft):
                    filled_output.extend(line)
                    if remaining := inner_width - fragment_list_width(line):
                        filled_output.append((theme["style"], (" " * remaining)))
                        # filled_output.append((theme["style"], ("·" * remaining)))
                    if filled_output and not filled_output[-1][1].endswith("\n"):
                        filled_output.append((theme["style"], "\n"))
                ft = filled_output

        if theme.get("invisible"):
            new_ft = []
            for line in split_lines(ft):
                new_ft.append((parent_theme["style"], " " * fragment_list_width(line)))
                new_ft.append(("", "\n"))
            if new_ft:
                new_ft.pop()
            ft = new_ft

        return ft

    def render_text(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Render a text element.

        Args:
            element: The page element to render
            parent_theme: The theme of the element's parent element
            available_width: The width available for rendering the element
            available_height: The height available for rendering the element
            left: The position on the current line at which to render the output - used
                to indent subsequent lines when rendering inline blocks like images
            preformatted: When True, whitespace in the the element's text is not
                collapsed

        Returns:
            Formatted text

        """
        ft: "StyleAndTextTuples" = []
        text = element.text
        style = parent_theme["style"]

        # Ensure hidden text is blank and not underlined
        if parent_theme.get("hidden"):
            text = " " * len(text)
            style = f"{style} nounderline"

        # Strip whitespace
        if not preformatted:
            strippable = True
            for i in text:
                if i not in "\x20\x0a\x09\x0c\x0d":
                    strippable = False
                    break
            if strippable:
                if "\n" in text:
                    text = "\n"
                else:
                    text = " "
            text = re.sub(r"\s+", " ", text.strip("\n").replace("\n", " "))

        if text:
            ft = [(style, text)]

        # Wrap non-pre-formatted text
        if not preformatted:
            ft = wrap(ft, available_width, left=left, style=style)

        return ft

    def render_li(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Render a list item."""
        # Get the element's theme
        theme = self.element_theme(element, parent_theme)
        # Get margin
        list_style = theme.get("list-style", "none")
        margin = {
            "none": "",
            "disc": "•",
            "circle": "○",
            "square": "■",
            "disclosure-open": "▼",
            "disclosure-closed": "▶",
        }.get(list_style, list_style)
        # Pad decimal bullets
        if list_style == "decimal":
            decimal = str(element.attrs["data-list-index"]).rjust(
                len(str(element.attrs["data-list-length"])), " "
            )
            margin = f"{decimal}."
        # Add space around bullet
        margin = f" {margin} "
        margin_len = len(margin)
        # Render the list item
        ft = self.render_contents(
            element.contents,
            parent_theme=theme,
            # Restrict the available width by the margin width
            available_width=available_width - margin_len,
            available_height=available_height,
            left=left,  # + margin_len,
            preformatted=preformatted,
        )
        # Indent using margin
        ft = [(f"{theme['style']} class:bullet", margin), *ft]
        ft = indent(ft, margin=" " * margin_len, skip_first=True, style=theme["style"])
        return ft

    def _render_li(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Render a list element."""
        # Get the element's theme
        theme = self.element_theme(element, parent_theme)
        # Get the bullet details
        bullet = str(element.attrs.get("data-margin", "•"))
        bullet_str = f" {bullet} "
        bullet_width = len(bullet_str)
        # Render the contents of the list item
        ft = self.render_contents(
            element.contents,
            parent_theme=theme,
            # Restrict the available width by the margin width
            available_width=available_width - bullet_width,
            available_height=available_height,
            left=left,
            preformatted=preformatted,
        )
        # Wrap the list item
        ft = wrap(
            ft,
            available_width - bullet_width,
            style=theme["style"],
            strip_trailing_ws=True,
        )
        # Add the bullet
        ft = [(f"{theme['style']} class:bullet", bullet_str), *ft]
        # Indent subsequent lines
        ft = indent(
            ft, margin=" " * bullet_width, skip_first=True, style=theme["style"]
        )
        ft = strip(ft, left=False, right=True)
        ft.append(("", "\n"))
        return ft

    def render_table(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Render a list of parsed markdown elements representing a table element.

        Args:
            element: The list of parsed elements to render
            available_width: The width available for rendering the element
            available_height: The height available for rendering the element
            parent_theme: The theme of the parent element
            left: The position on the current line at which to render the output - used
                to indent subsequent lines when rendering inline blocks like images
            preformatted: When True, whitespace in the the element's text is not
                collapsed

        Returns:
            Formatted text

        """
        ft = []

        table_theme = self.element_theme(element, parent_theme)

        # Adjust the width
        inner_width = available_width
        if (theme_width := table_theme.get("width")) is not None:
            if isinstance(theme_width, float) and 0 < theme_width <= 1:
                inner_width = int(inner_width * theme_width + 0.5)
            else:
                inner_width = min(inner_width, theme_width)

        if margin := table_theme.get("margin"):
            for margin_size in (margin.left, margin.right):
                if isinstance(margin_size, float) and 0 < margin_size <= 1:
                    available_width = int(inner_width * (1 - margin_size))
                else:
                    available_width -= margin_size or 0

        if padding := table_theme.get("padding"):
            for padding_size in (padding.left, padding.right):
                if isinstance(padding_size, float) and 0 < padding_size <= 1:
                    inner_width = int(inner_width * (1 - padding_size))
                else:
                    inner_width -= padding_size or 0

        table = Table(
            align=table_theme["align"],
            style=table_theme["style"],
            padding=table_theme["padding"],
            border_style=table_theme["style"],
            # Use invisible borders by default
            border=table_theme["border"] or Invisible,
            border_collapse=table_theme["border_collapse"],
        )

        # Stack the elements in the shape of the table
        def render_rows(elements: "list[PageElement]") -> "None":
            for tr in elements:
                if tr.name == "tr":
                    tr_theme = self.element_theme(tr, table_theme)
                    row = table.new_row(
                        align=tr_theme["align"],
                        padding=tr_theme["padding"],
                        border=tr_theme["border"],
                        style=tr_theme["style"],
                    )
                    for td in tr.contents:
                        if td.name in ("th", "td"):
                            td_theme = self.element_theme(td, tr_theme)
                            row.new_cell(
                                text=self.render_contents(
                                    td.contents,
                                    parent_theme=td_theme,
                                    available_width=available_width,
                                    available_height=available_height,
                                    left=0,
                                    preformatted=preformatted,
                                ),
                                padding=td_theme["padding"],
                                border=td_theme["border"],
                                align=td_theme["align"],
                                colspan=try_eval(td.attrs.get("colspan", 1)),
                                rowspan=try_eval(td.attrs.get("rowspan", 1)),
                                style=td_theme["style"] + " nounderline",
                            )

        # Render the table head
        for child in element.find_all("thead", recursive=False):
            render_rows(child.contents)
        # Render the table body
        for child in element.find_all("tbody", recursive=False):
            render_rows(child.contents)
        # Render rows not in a head / body / foot as part of the body
        render_rows(element.contents)
        for child in element.find_all("tfoot", recursive=False):
            render_rows(child.contents)
        # Render the table
        ft_table = table.render(inner_width)

        # Render the caption
        captions = element.find_all("caption", recursive=False)
        if captions:
            table_width = max_line_width(ft_table)
            for child in captions:
                ft_caption = self.render_element(
                    child, table_theme, table_width, left, preformatted
                )
                if ft_caption:
                    ft.extend(ft_caption)

        ft.extend(ft_table)

        return ft

    def render_details(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Render an expand summary / details."""
        ft: "StyleAndTextTuples" = []
        theme = self.element_theme(element, parent_theme)

        summary_element = None
        for child in element.contents:
            if child.name == "summary":
                summary_element = child
                break

        if summary_element is not None:
            ft.extend(
                self.render_element(
                    summary_element,
                    parent_theme=theme,
                    available_width=available_width,
                    available_height=available_height,
                    left=left,
                    preformatted=preformatted,
                )
            )
            ft.append(("", "\n"))

        detail_elements = [e for e in element.contents if e is not summary_element]

        ft.extend(
            self.format_details(
                self.render_contents(
                    detail_elements,
                    parent_theme=theme,
                    available_width=available_width - _ELEMENT_INSETS["details"],
                    available_height=available_height,
                    left=left,
                    preformatted=preformatted,
                ),
                available_width,
                available_height,
                left,
                element,
                theme,
            )
        )
        ft.append(("", "\n"))
        return ft

    def _render_image(
        self,
        data: "Any",
        format_: "str",
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
    ) -> "Optional[StyleAndTextTuples]":
        cols, aspect = pixels_to_cell_size(*data_pixel_size(data, format_=format_))
        # Manially set a value if we don't have one
        cols = cols or 20
        aspect = aspect or 0.5
        # Scale down the image to fit to width
        cols = min(available_width, cols)
        rows = ceil(cols * aspect)
        # Convert the image to formatted-text
        result = convert(
            data,
            format_,
            "formatted_text",
            cols=cols,
            rows=rows,
        )
        # Remove trailing new-lines
        result = strip(result, char="\n", left=False)
        # Optionally add a border
        if try_eval(element.attrs.get("border", "")):
            result = add_border(
                result,
                border=Rounded.grid,
                style="class:md.img.border",
            )
        # Indent for line continuation as images are inline
        # if left:
        result = indent(
            result, " " * left, skip_first=True, style=parent_theme["style"]
        )
        return result

    def render_img(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Display images rendered as ANSI art."""
        src = str(element.attrs.get("src", ""))
        # Attempt to load the image url
        if data := load_url(src, self.base):
            # Display it graphically
            format_ = get_format(src, default="png")
            ft = self._render_image(
                data,
                format_,
                element,
                parent_theme,
                available_width,
                available_height,
                left,
            )
            return ft
        # Otherwise, display the image title
        else:
            return self.render_element(
                element,
                parent_theme,
                available_width,
                available_height,
                left,
                preformatted,
            )

    def render_svg(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Display images rendered as ANSI art."""
        # Ensure xml namespace is set
        element.attrs.setdefault("xmlns", "http://www.w3.org/2000/svg")
        element.attrs.setdefault("xmlns:xlink", "http://www.w3.org/1999/xlink")
        data = element._outer_html()
        return self._render_image(
            data, "svg", element, parent_theme, available_width, available_height, left
        )

    def render_math(
        self,
        element: "PageElement",
        parent_theme: "dict[str, Any]",
        available_width: "int",
        available_height: "int",
        left: "int" = 0,
        preformatted: "bool" = False,
    ) -> "StyleAndTextTuples":
        """Display LaTeX maths rendered as unicode text."""
        text = "".join(desc.text for desc in element.descendents)
        for desc in element.descendents:
            if desc.name == "text":
                desc.text = convert(text, "latex", "ansi")
        return self.render_element(
            element,
            parent_theme,
            available_width,
            available_height,
            left,
            preformatted=True,
        )

    ###

    def __pt_formatted_text__(self) -> "StyleAndTextTuples":
        """Formatted text magic method."""
        return self.formatted_text

    # Tag formatting methods

    @staticmethod
    def format_img(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Fallback to formatting an image as title text if nothing loaded."""
        if not ft:
            bounds = ("", "")
            # Add fallback text if there is no image title
            title = str(element.attrs.get("alt"))
            # Try getting the filename
            if not title:
                src = str(element.attrs.get("src", ""))
                if not src.startswith("data:"):
                    title = src.rsplit("/", 1)[-1]
            if not title:
                title = "Image"
            # Add the sunrise emoji to represent an image. I would use :framed_picture:, but it
            # requires multiple code-points and causes breakage in many terminals
            ft = [
                ("reverse", f"{bounds[0]}"),
                ("", f"🌄 {title}"),
                ("reverse", f"{bounds[1]}"),
            ]
            ft = apply_style(ft, theme["style"])
        return ft

    @staticmethod
    def format_div(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format a horizontal rule."""
        # ft = wrap(ft, available_width, style=theme["style"])
        return ft

    @staticmethod
    def format_hr(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format a horizontal rule."""
        ft = [(theme["style"], "─" * available_width)]
        return ft

    @staticmethod
    def format_h1(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format a top-level heading wrapped and centered with a full width double border."""
        ft = wrap(ft, available_width - 4, style=theme["style"])
        ft = align(
            FormattedTextAlign.CENTER,
            ft,
            width=available_width - 4,
            style=theme["style"],
        )
        ft = add_border(
            ft,
            width=available_width,
            border=Double.grid,
            style=theme["style"],
        )
        return ft

    @staticmethod
    def format_h2(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format a 2nd-level headding wrapped and centered with a double border."""
        ft = wrap(ft, width=available_width - 4, style=theme["style"])
        ft = align(FormattedTextAlign.CENTER, ft)
        ft = add_border(
            ft,
            border=Thin.grid,
            style=theme["style"],
        )
        ft = align(
            FormattedTextAlign.CENTER, ft, width=available_width, style=theme["style"]
        )
        return ft

    @staticmethod
    def format_h(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format headings wrapped and centered."""
        ft = wrap(ft, available_width, style=theme["style"])
        ft = align(FormattedTextAlign.CENTER, ft, width=available_width)
        return ft

    @staticmethod
    def format_p(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format paragraphs wrapped."""
        ft = wrap(ft, available_width, style=theme["style"])
        return ft

    @staticmethod
    def format_br(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format line breaks."""
        return [("", "\n")]

    @staticmethod
    def format_blockquote(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format blockquotes with a solid left margin."""
        ft = wrap(strip(ft, char="\n"), width=available_width, style=theme["style"])
        ft = indent(ft, margin="▌ ", style=f"{theme['style']} class:margin")
        return ft

    @staticmethod
    def format_pre(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format a pre-formatted block with a border if it contains code."""
        if element.contents and element.contents[0].name == "code":
            ft = align(FormattedTextAlign.LEFT, ft, available_width - 4)
            ft = add_border(ft, available_width, border=Thin.grid)
        return ft

    @staticmethod
    def format_a(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format hyperlinks and adds link escape sequences."""
        href = element.attrs.get("href")
        if href:
            link_id = randint(0, 999999)  # noqa S311
            result = []
            for line in split_lines(ft):
                result += [
                    ("[ZeroWidthEscape]", tmuxify(f"\x1b]8;id={link_id};{href}\x1b\\")),
                    *line,
                    ("[ZeroWidthEscape]", tmuxify("\x1b]8;;\x1b\\")),
                    ("", "\n"),
                ]
            result.pop()
            return result
        else:
            return ft

    @staticmethod
    def format_summary(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format summaries with an arrow."""
        return [
            ("class:html.summary.arrow", " ⮟ "),
            *indent(
                wrap(ft, width=available_width, style=theme["style"]),
                margin="   ",
                skip_first=True,
                style=theme["style"],
            ),
        ]

    @staticmethod
    def format_details(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format details indented."""
        ft = strip(ft, char="\n")
        ft = indent(ft, margin="   ")
        return ft

    @staticmethod
    def format_q(
        ft: "StyleAndTextTuples",
        available_width: "int",
        available_height: "int",
        left: "int",
        element: "PageElement",
        theme: "dict",
    ) -> "StyleAndTextTuples":
        """Format an inline quote."""
        return [("", '"'), *ft, ("", '"')]


if __name__ == "__main__":
    import sys

    from prompt_toolkit.application.current import create_app_session, set_app
    from prompt_toolkit.shortcuts.utils import print_formatted_text
    from prompt_toolkit.styles.style import Style
    from upath import UPath

    from euporie.core.app import BaseApp
    from euporie.core.style import HTML_STYLE

    path = UPath(sys.argv[1])

    with create_app_session(input=BaseApp.load_input(), output=BaseApp.load_output()):
        with set_app(BaseApp()):
            with path.open() as f:
                print_formatted_text(
                    HTML(path.open().read(), base=path, pad=True),
                    style=Style(HTML_STYLE),
                )



"""Defines default key-bindings."""

from __future__ import annotations

from typing import TYPE_CHECKING

from prompt_toolkit.key_binding import KeyBindings

from euporie.core.commands import get_cmd
from euporie.core.config import Config
from euporie.core.current import get_app
from euporie.core.keys import Keys

if TYPE_CHECKING:

    from typing import Dict, List, Optional, Tuple, Union

    from prompt_toolkit.key_binding import KeyBindingsBase

    AnyKey = Union[Tuple[Union[Keys, str], ...], Keys, str]
    AnyKeys = Union[List[AnyKey], AnyKey]

    KeyBindingDefs = Dict[str, AnyKeys]

BINDINGS: "dict[str, KeyBindingDefs]" = {}


def register_bindings(bindings: "dict[str, KeyBindingDefs]") -> "None":
    """Updates the key-binding registry."""
    for group, items in bindings.items():
        if group not in BINDINGS:
            BINDINGS[group] = {}
        for command, keys in items.items():
            BINDINGS[group][command] = keys


def load_registered_bindings(
    *names: "str", config: "Optional[Config]" = None
) -> "KeyBindingsBase":
    """Assign key-bindings to commands based on a dictionary."""
    from euporie.core.app import BaseApp

    app = get_app()
    if config is None and isinstance(app, BaseApp):
        config = app.config

    kb = KeyBindings()
    for name in names:
        binding_dict = BINDINGS.get(name, {})
        # Augment with bindings from config
        if config is not None:
            binding_dict.update(config.key_bindings.get(name, {}))
        for command, keys in binding_dict.items():
            get_cmd(command).bind(kb, keys)
    return kb



"""Contains the command palette container."""

from __future__ import annotations

import logging
import re
from functools import partial
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.layout.containers import HSplit, ScrollOffsets, VSplit, Window
from prompt_toolkit.layout.controls import UIContent, UIControl
from prompt_toolkit.layout.margins import ScrollbarMargin
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType

from euporie.core.app import get_app
from euporie.core.commands import Command, add_cmd, commands
from euporie.core.key_binding.registry import register_bindings
from euporie.core.widgets.dialog import Dialog
from euporie.core.widgets.forms import Text

if TYPE_CHECKING:
    from typing import Optional

    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.controls import GetLinePrefixCallable

    from euporie.core.app import BaseApp, StatusBarFields

log = logging.getLogger(__name__)


_CommandMatch = NamedTuple(
    "_CommandMatch",
    [("length", int), ("position", int), ("command", Command)],
)


class CommandMenuControl(UIControl):
    """A filterable and navigable list of commands."""

    def __init__(self, command_palette: "CommandPalette", width: "int" = 60) -> "None":
        """Instantiates a new command list instance."""
        self.palette = command_palette
        self.width = width

    # def is_focusable(self) -> "bool":
    # return True

    def create_content(self, width: "int", height: "int") -> "UIContent":
        """Create a UIContent object for this control."""

        def get_line(i: "int") -> "StyleAndTextTuples":
            result: "StyleAndTextTuples" = []
            match = self.palette.matches[i]

            # Set style depending on row number / selection status
            if i == self.palette.index:
                style = "class:palette.item.selected"
            elif (i + 1) % 2 == 0:
                style = "class:palette.item.alt"
            else:
                style = "class:palette.item"

            # Calculate prefix
            if match.command.toggled is not None and match.command.toggled():
                prefix = " ✓ "
            else:
                prefix = "   "

            # Calculate suffix
            suffix = match.command.key_str() + " "

            # Add prefix with status for toggle commands
            result.append((style, prefix))

            # Get list of input character and search text
            chars = list(self.palette.text_area.text)
            text = match.command.title.ljust(width - len(prefix) - len(suffix))
            # Add text before the match
            result.append((style, text[: match.position]))
            # Format text, undelining matching characters
            for c in text[match.position : match.position + match.length]:
                if chars and c.lower() == chars[0].lower():
                    result.append((style + " underline", c))
                    del chars[0]
                else:
                    result.append((style, c))
            # Add text after the match
            result.append((style, text[match.position + match.length :]))

            # Add suffix with key-binding
            result.append((style + " italic", suffix))

            return result

        return UIContent(
            get_line=get_line,
            cursor_position=Point(x=0, y=self.palette.index * 1 or 0),
            line_count=len(self.palette.matches),
        )

    def preferred_width(self, max_available_width: "int") -> "Optional[int]":
        """Returns the preferred width of the command list."""
        return min(self.width, max_available_width)

    def preferred_height(
        self,
        width: "int",
        max_available_height: "int",
        wrap_lines: "bool",
        get_line_prefix: "Optional[GetLinePrefixCallable]",
    ) -> "Optional[int]":
        """Returns the preferred height of the command list."""
        return min(
            max_available_height // 2,
            len(self.palette.matches),
        )

    def mouse_handler(self, mouse_event: "MouseEvent") -> "NotImplementedOrNone":
        """Handle clicking and scrolling mouse events."""
        if mouse_event.event_type == MouseEventType.MOUSE_UP:
            self.palette.index = mouse_event.position.y
            self.palette.accept()

        if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            # Scroll up.
            self.palette.select(1)

        elif mouse_event.event_type == MouseEventType.SCROLL_UP:
            # Scroll down.
            self.palette.select(-1)

        return None


class CommandPalette(Dialog):
    """A command palette which allows searching the available commands."""

    index: "int"
    matches: "list[_CommandMatch]"

    title = "Search for a command"
    body_padding_top = 0
    body_padding_bottom = 0

    def __init__(self, app: "BaseApp") -> None:
        """Instantiates a new command palette instance."""
        super().__init__(app=app)
        self.matches: "list[_CommandMatch]" = []
        self.index = 0

        self.last_focused: "Optional[UIControl]" = None
        self._visible = False
        self.visible = Condition(lambda: self._visible)

        # self.kb = KeyBindings()
        self.kb.add("s-tab")(focus_previous)
        self.kb.add("tab")(focus_next)
        self.kb.add("escape")(self.hide)
        self.kb.add("up", filter=Condition(lambda: bool(self.matches)))(
            partial(self.select, -1)
        )
        self.kb.add("down", filter=Condition(lambda: bool(self.matches)))(
            partial(self.select, 1)
        )
        self.kb.add("pageup", filter=Condition(lambda: bool(self.matches)))(
            partial(self.select, -10)
        )
        self.kb.add("pagedown", filter=Condition(lambda: bool(self.matches)))(
            partial(self.select, 10)
        )
        self.kb.add("home", filter=Condition(lambda: bool(self.matches)))(
            partial(lambda e: self.select(-self.index))
        )
        self.kb.add("end", filter=Condition(lambda: bool(self.matches)))(
            partial(lambda e: self.select(len(self.matches)))
        )

        self.text_area = Text(
            multiline=False,
            accept_handler=self.accept,
            style="class:input",
            expand=False,
            placeholder="  Type to search…",
        )
        self.text_area.buffer.on_text_changed += self.text_changed
        scroll_bar_margin = ScrollbarMargin(display_arrows=True)
        scroll_bar_margin.up_arrow_symbol = "▲"
        scroll_bar_margin.down_arrow_symbol = "▼"

        self.body = HSplit(
            [
                VSplit(
                    [self.text_area],
                    padding=1,
                ),
                Window(
                    CommandMenuControl(self),
                    scroll_offsets=ScrollOffsets(bottom=1),
                    right_margins=[scroll_bar_margin],
                ),
            ],
        )
        self.buttons = {}

        get_app().container_statuses[self.container] = self.statusbar_fields

    def load(self) -> "None":
        """The body is already loaded: does nothing."""
        self.text_area.buffer.text = ""
        self.to_focus = self.text_area

    def statusbar_fields(
        self,
    ) -> "StatusBarFields":
        """Returns a list of statusbar field values shown then this tab is active."""
        if self.matches:
            return ([self.matches[self.index].command.description], [])
        else:
            return ([], [])

    def select(self, n: "int", event: "KeyPressEvent" = None) -> "None":
        """Change the index of the selected command.

        Args:
            n: The relative amount by which to change the selected index
            event: Ignored

        """
        self.index += n
        self.index = min(len(self.matches) - 1, max(0, self.index))

    def text_changed(self, buffer: "Buffer") -> "None":
        """Called when the input text changes: filters the command list."""
        self.matches = []

        # Lookahead regex to manage overlapping matches
        regex = re.compile(
            "(?=({0}))".format(".*?".join(map(re.escape, buffer.text))), re.IGNORECASE
        )
        for cmd in commands.values():
            if not cmd.hidden():
                matches = list(regex.finditer(cmd.title))
                if matches:
                    # Prefer the match closest to the left, then shortest.
                    best = min(matches, key=lambda m: (len(m.group(1)), m.start()))
                    self.matches.append(
                        _CommandMatch(len(best.group(1)), best.start(), cmd)
                    )
        self.matches = sorted(self.matches, key=lambda m: (m.length, m.position))

        # Ensure the selected index is within the list of matches
        self.index = min(len(self.matches) - 1, max(0, self.index))

    def accept(self, buffer: "Optional[Buffer]" = None) -> "bool":
        """Called on :kbd:`enter`: runs the selected command."""
        if self.matches:
            self.hide()
            self.matches[self.index].command.run()
            return True
        else:
            return False

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd()
    def _toggle_command_palette() -> "None":
        """Show the command palette."""
        if command_palette := get_app().dialogs.get("command-palette"):
            command_palette.toggle()

    @staticmethod
    @add_cmd()
    def _show_command_palette() -> "None":
        """Show the command palette."""
        if command_palette := get_app().dialogs.get("command-palette"):
            command_palette.show()

    @staticmethod
    @add_cmd()
    def _hide_command_palette() -> "None":
        """Hide the command palette."""
        if command_palette := get_app().dialogs.get("command-palette"):
            command_palette.hide()

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.app.BaseApp": {
                "show-command-palette": "c-@",
            }
        }
    )



"""Defines border styles."""

from __future__ import annotations

from enum import Enum
from functools import lru_cache, total_ordering
from typing import TYPE_CHECKING, NamedTuple

if TYPE_CHECKING:
    from typing import Optional


class GridPart(Enum):
    """Defines the component characters of a grid.

    Character naming works as follows:

                ╭┈┈┈┈┈┈┈┈LEFT
                ┊ ╭┈┈┈┈┈┈MID
                ┊ ┊ ╭┈┈┈┈SPLIT
                ┊ ┊ ┊ ╭┈┈RIGHT
                ∨ ∨ ∨ v
          TOP┈> ┏ ━ ┳ ┓
          MID┈> ┃   ┃ ┃
        SPLIT┈> ┣ ━ ╋ ┫
       BOTTOM┈> ┗ ━ ┻ ┛

    """

    TOP_LEFT = 0
    TOP_MID = 1
    TOP_SPLIT = 2
    TOP_RIGHT = 3
    MID_LEFT = 4
    MID_MID = 5
    MID_SPLIT = 6
    MID_RIGHT = 7
    SPLIT_LEFT = 8
    SPLIT_MID = 9
    SPLIT_SPLIT = 10
    SPLIT_RIGHT = 11
    BOTTOM_LEFT = 12
    BOTTOM_MID = 13
    BOTTOM_SPLIT = 14
    BOTTOM_RIGHT = 15


class DirectionFlags(NamedTuple):
    """Flags which indicate the connection of a grid node."""

    north: "bool" = False
    east: "bool" = False
    south: "bool" = False
    west: "bool" = False


class Mask:
    """A mask which allows selection of a subset of a grid.

    Masks can be combined to construct more complex masks.
    """

    def __init__(self, mask: "dict[GridPart, DirectionFlags]") -> "None":
        """Create a new grid mask.

        Args:
            mask: A dictionary mapping grid parts to a tuple of direction flags.
                Not all :class:`GridPart`s need to be defined - any which are not
                defined are assumed to be set entirely to :cont:`False`.

        """
        self.mask = {part: mask.get(part, DirectionFlags()) for part in GridPart}

    def __add__(self, other: "Mask") -> "Mask":
        """Adds two masks, combining direction flags for each :class:`GridPart`.

        Args:
            other: Another mask to combine with this one

        Returns:
            A new combined mask.

        """
        return Mask(
            {
                key: DirectionFlags(
                    *(self.mask[key][i] | other.mask[key][i] for i in range(4))
                )
                for key in GridPart
            }
        )


class Masks:
    """A collection of default masks."""

    top_edge = Mask(
        {
            GridPart.TOP_LEFT: DirectionFlags(False, True, False, False),
            GridPart.TOP_MID: DirectionFlags(False, True, False, True),
            GridPart.TOP_SPLIT: DirectionFlags(False, True, False, True),
            GridPart.TOP_RIGHT: DirectionFlags(False, False, False, True),
        }
    )

    middle_edge = Mask(
        {
            GridPart.SPLIT_LEFT: DirectionFlags(False, True, False, False),
            GridPart.SPLIT_MID: DirectionFlags(False, True, False, True),
            GridPart.SPLIT_SPLIT: DirectionFlags(False, True, False, True),
            GridPart.SPLIT_RIGHT: DirectionFlags(False, False, False, True),
        }
    )

    bottom_edge = Mask(
        {
            GridPart.BOTTOM_LEFT: DirectionFlags(False, True, False, False),
            GridPart.BOTTOM_MID: DirectionFlags(False, True, False, True),
            GridPart.BOTTOM_SPLIT: DirectionFlags(False, True, False, True),
            GridPart.BOTTOM_RIGHT: DirectionFlags(False, False, False, True),
        }
    )

    left_edge = Mask(
        {
            GridPart.TOP_LEFT: DirectionFlags(False, False, True, False),
            GridPart.MID_LEFT: DirectionFlags(True, False, True, False),
            GridPart.SPLIT_LEFT: DirectionFlags(True, False, True, False),
            GridPart.BOTTOM_LEFT: DirectionFlags(True, False, False, False),
        }
    )

    center_edge = Mask(
        {
            GridPart.TOP_SPLIT: DirectionFlags(False, False, True, False),
            GridPart.MID_SPLIT: DirectionFlags(True, False, True, False),
            GridPart.SPLIT_SPLIT: DirectionFlags(True, False, True, False),
            GridPart.BOTTOM_SPLIT: DirectionFlags(True, False, False, False),
        }
    )

    right_edge = Mask(
        {
            GridPart.TOP_RIGHT: DirectionFlags(False, False, True, False),
            GridPart.MID_RIGHT: DirectionFlags(True, False, True, False),
            GridPart.SPLIT_RIGHT: DirectionFlags(True, False, True, False),
            GridPart.BOTTOM_RIGHT: DirectionFlags(True, False, False, False),
        }
    )

    inner = center_edge + middle_edge
    outer = top_edge + right_edge + bottom_edge + left_edge

    grid = inner + outer

    corners = Mask(
        {
            GridPart.TOP_LEFT: DirectionFlags(False, True, True, False),
            GridPart.TOP_RIGHT: DirectionFlags(False, False, True, True),
            GridPart.BOTTOM_LEFT: DirectionFlags(True, True, False, False),
            GridPart.BOTTOM_RIGHT: DirectionFlags(True, False, False, True),
        }
    )


@total_ordering
class LineStyle:
    """Defines a line style which can be used to draw grids.

    :class:`GridStyle`s can be created from a :class:`LineStyle` by accessing an
    attribute with the name of a default mask from :class:`Masks`.
    """

    def __init__(
        self, name: "str", rank: "int", parent: "Optional[LineStyle]" = None
    ) -> "None":
        """Creates a new :class:`LineStyle`.

        Args:
            name: The name of the line style
            rank: A ranking value - this is used when two adjoining cells in a table
                have differing line styles to determine which should take precedence.
                The style with the higher rank is used.
            parent: The :class:`LineStyle` from which this style should inherit if any
                characters are undefined.

        """
        self.name = name
        self.rank = rank
        self.children: "dict[str, LineStyle]" = {}
        self.parent = parent
        if parent:
            parent.children[name] = self

    def __getattr__(self, value: "str") -> "GridStyle":
        """Defines attribute access.

        Allows dynamic access to children and :class:`GridStyle` creation via attribute
        access.

        Args:
            value: The attribute name to access

        Returns:
            The accessed attribute

        Raises:
            AttributeError: Raised if there is no such attribute

        """
        if hasattr(Masks, value):
            mask = getattr(Masks, value)
            return GridStyle(self, mask)
        else:
            raise AttributeError(f"No such attribute `{value}`")

    def __dir__(self) -> "list[str]":
        """Lists the public attributes."""
        return [x for x in Masks.__dict__ if not x.startswith("_")]

    def __lt__(self, other: "LineStyle") -> "bool":
        """Allows :class:`LineStyle`s to be sorted according to their rank."""
        if self.__class__ is other.__class__:
            return self.rank < other.rank
        elif other is None:
            return None
        return NotImplemented

    def __repr__(self) -> "str":
        """A string representation of the :class:`LineStyle`."""
        return f"LineStyle({self.name})"


# Line Styles
Invisible = LineStyle("Invisible", rank=0)
Ascii = LineStyle("Ascii", rank=10)
AsciiThick = LineStyle("AsciiDouble", rank=11, parent=Ascii)
Thin = LineStyle("Thin", rank=20, parent=Ascii)
Thick = LineStyle("Thick", rank=30, parent=Ascii)
Rounded = LineStyle("Rounded", rank=40, parent=Thin)
ThinDoubleDashed = LineStyle("DoubleDashed", 51, parent=Thin)
ThinTripleDashed = LineStyle("TripleDashed", 52, parent=Thin)
ThinQuadrupleDashed = LineStyle("QuadDashed", 53, parent=Thin)
ThickDoubleDashed = LineStyle("DoubleDashed", 54, parent=Thick)
ThickTripleDashed = LineStyle("TripleDashed", 35, parent=Thick)
ThickQuadrupleDashed = LineStyle("QuadDashed", 56, parent=Thick)
Double = LineStyle("Double", 50, parent=Thick)
EighthBlockUpperRight = LineStyle("EighthBlockUpperRight", 50, parent=Thin)
EighthBlockLowerLeft = LineStyle("EighthBlockLowerLeft", 50, parent=Thin)
HalfBlockUpperRight = LineStyle("HalfBlockUpperRight", 50, parent=Thin)
HalfBlockLowerLeft = LineStyle("HalfBlockLowerLeft", 50, parent=Thin)


class GridChar(NamedTuple):
    """Representation of a grid node character.

    The four compass points represent the line style joining from the given direction.
    """

    north: "LineStyle"
    east: "LineStyle"
    south: "LineStyle"
    west: "LineStyle"


_GRID_CHARS = {
    # NONE
    GridChar(Invisible, Invisible, Invisible, Invisible): " ",
    # Ascii
    GridChar(Ascii, Invisible, Ascii, Invisible): "|",
    GridChar(Invisible, Ascii, Invisible, Ascii): "-",
    GridChar(Ascii, Ascii, Invisible, Invisible): "+",
    GridChar(Invisible, Ascii, Ascii, Invisible): "+",
    GridChar(Invisible, Invisible, Ascii, Ascii): "+",
    GridChar(Ascii, Invisible, Invisible, Ascii): "+",
    GridChar(Ascii, Ascii, Ascii, Invisible): "+",
    GridChar(Invisible, Ascii, Ascii, Ascii): "+",
    GridChar(Ascii, Invisible, Ascii, Ascii): "+",
    GridChar(Ascii, Ascii, Invisible, Ascii): "+",
    GridChar(Ascii, Ascii, Ascii, Ascii): "+",
    # AsciiThick
    GridChar(Invisible, AsciiThick, Invisible, AsciiThick): "=",
    # Thin
    GridChar(Thin, Invisible, Invisible, Invisible): "╵",
    GridChar(Invisible, Thin, Invisible, Invisible): "╶",
    GridChar(Invisible, Invisible, Thin, Invisible): "╷",
    GridChar(Invisible, Invisible, Invisible, Thin): "╴",
    GridChar(Thin, Invisible, Thin, Invisible): "│",
    GridChar(Invisible, Thin, Invisible, Thin): "─",
    GridChar(Thin, Thin, Invisible, Invisible): "└",
    GridChar(Invisible, Thin, Thin, Invisible): "┌",
    GridChar(Invisible, Invisible, Thin, Thin): "┐",
    GridChar(Thin, Invisible, Invisible, Thin): "┘",
    GridChar(Thin, Thin, Thin, Invisible): "├",
    GridChar(Invisible, Thin, Thin, Thin): "┬",
    GridChar(Thin, Invisible, Thin, Thin): "┤",
    GridChar(Thin, Thin, Invisible, Thin): "┴",
    GridChar(Thin, Thin, Thin, Thin): "┼",
    # Thin Rounded
    GridChar(Rounded, Rounded, Invisible, Invisible): "╰",
    GridChar(Invisible, Rounded, Rounded, Invisible): "╭",
    GridChar(Invisible, Invisible, Rounded, Rounded): "╮",
    GridChar(Rounded, Invisible, Invisible, Rounded): "╯",
    # Thin Dashes
    GridChar(Invisible, ThinDoubleDashed, Invisible, ThinDoubleDashed): "╌",
    GridChar(ThinDoubleDashed, Invisible, ThinDoubleDashed, Invisible): "╎",
    GridChar(Invisible, ThinTripleDashed, Invisible, ThinTripleDashed): "┄",
    GridChar(ThinTripleDashed, Invisible, ThinTripleDashed, Invisible): "┆",
    GridChar(Invisible, ThinQuadrupleDashed, Invisible, ThinQuadrupleDashed): "┈",
    GridChar(ThinQuadrupleDashed, Invisible, ThinQuadrupleDashed, Invisible): "┊",
    # Double
    GridChar(Double, Invisible, Invisible, Invisible): "║",
    GridChar(Invisible, Double, Invisible, Invisible): "═",
    GridChar(Invisible, Invisible, Double, Invisible): "║",
    GridChar(Invisible, Invisible, Invisible, Double): "═",
    GridChar(Double, Invisible, Double, Invisible): "║",
    GridChar(Invisible, Double, Invisible, Double): "═",
    GridChar(Double, Double, Invisible, Invisible): "╚",
    GridChar(Invisible, Double, Double, Invisible): "╔",
    GridChar(Invisible, Invisible, Double, Double): "╗",
    GridChar(Double, Invisible, Invisible, Double): "╝",
    GridChar(Double, Double, Double, Invisible): "╠",
    GridChar(Invisible, Double, Double, Double): "╦",
    GridChar(Double, Invisible, Double, Double): "╣",
    GridChar(Double, Double, Invisible, Double): "╩",
    GridChar(Double, Double, Double, Double): "╬",
    # Double / Thin
    GridChar(Thin, Double, Thin, Double): "╪",
    GridChar(Double, Thin, Double, Thin): "╫",
    GridChar(Double, Thin, Invisible, Invisible): "╙",
    GridChar(Invisible, Double, Thin, Invisible): "╒",
    GridChar(Invisible, Invisible, Double, Thin): "╖",
    GridChar(Thin, Invisible, Invisible, Double): "╛",
    GridChar(Double, Invisible, Invisible, Thin): "╜",
    GridChar(Thin, Double, Invisible, Invisible): "╘",
    GridChar(Invisible, Thin, Double, Invisible): "╓",
    GridChar(Invisible, Invisible, Thin, Double): "╕",
    GridChar(Thin, Double, Thin, Invisible): "╞",
    GridChar(Invisible, Thin, Double, Thin): "╥",
    GridChar(Thin, Invisible, Thin, Double): "╡",
    GridChar(Double, Thin, Invisible, Thin): "╨",
    GridChar(Double, Thin, Double, Invisible): "╟",
    GridChar(Invisible, Double, Thin, Double): "╤",
    GridChar(Double, Invisible, Double, Thin): "╢",
    GridChar(Thin, Double, Invisible, Double): "╧",
    # Thick
    GridChar(Thick, Thick, Thick, Thick): "╋",
    GridChar(Thick, Invisible, Invisible, Invisible): "╹",
    GridChar(Invisible, Thick, Invisible, Invisible): "╺",
    GridChar(Invisible, Invisible, Thick, Invisible): "╻",
    GridChar(Invisible, Invisible, Invisible, Thick): "╸",
    GridChar(Thick, Invisible, Thick, Invisible): "┃",
    GridChar(Invisible, Thick, Invisible, Thick): "━",
    GridChar(Thick, Thick, Invisible, Invisible): "┗",
    GridChar(Invisible, Thick, Thick, Invisible): "┏",
    GridChar(Invisible, Invisible, Thick, Thick): "┓",
    GridChar(Thick, Invisible, Invisible, Thick): "┛",
    GridChar(Thick, Thick, Thick, Invisible): "┣",
    GridChar(Invisible, Thick, Thick, Thick): "┳",
    GridChar(Thick, Invisible, Thick, Thick): "┫",
    GridChar(Thick, Thick, Invisible, Thick): "┻",
    # Thick Dashes
    GridChar(Invisible, ThickDoubleDashed, Invisible, ThickDoubleDashed): "╍",
    GridChar(ThickDoubleDashed, Invisible, ThickDoubleDashed, Invisible): "╏",
    GridChar(Invisible, ThickTripleDashed, Invisible, ThickTripleDashed): "┅",
    GridChar(ThickTripleDashed, Invisible, ThickTripleDashed, Invisible): "┇",
    GridChar(Invisible, ThickQuadrupleDashed, Invisible, ThickQuadrupleDashed): "┉",
    GridChar(ThickQuadrupleDashed, Invisible, ThickQuadrupleDashed, Invisible): "┋",
    # Thick / Thin
    GridChar(Invisible, Thick, Invisible, Thin): "╼",
    GridChar(Thin, Invisible, Thick, Invisible): "╽",
    GridChar(Invisible, Thin, Invisible, Thick): "╾",
    GridChar(Thick, Invisible, Thin, Invisible): "╿",
    GridChar(Thick, Thin, Invisible, Invisible): "┖",
    GridChar(Invisible, Thick, Thin, Invisible): "┍",
    GridChar(Invisible, Invisible, Thick, Thin): "┒",
    GridChar(Thin, Invisible, Invisible, Thick): "┙",
    GridChar(Thick, Invisible, Invisible, Thin): "┚",
    GridChar(Thin, Thick, Invisible, Invisible): "┕",
    GridChar(Invisible, Thin, Thick, Invisible): "┎",
    GridChar(Invisible, Invisible, Thin, Thick): "┑",
    GridChar(Thick, Thin, Thin, Invisible): "┞",
    GridChar(Invisible, Thick, Thin, Thin): "┮",
    GridChar(Thin, Invisible, Thick, Thin): "┧",
    GridChar(Thin, Thin, Invisible, Thick): "┵",
    GridChar(Thick, Invisible, Thin, Thin): "┦",
    GridChar(Thin, Thick, Invisible, Thin): "┶",
    GridChar(Thin, Thin, Thick, Invisible): "┟",
    GridChar(Invisible, Thin, Thin, Thick): "┭",
    GridChar(Thick, Thin, Invisible, Thin): "┸",
    GridChar(Thin, Thick, Thin, Invisible): "┝",
    GridChar(Invisible, Thin, Thick, Thin): "┰",
    GridChar(Thin, Invisible, Thin, Thick): "┥",
    GridChar(Thick, Thick, Thin, Invisible): "┡",
    GridChar(Invisible, Thick, Thick, Thin): "┲",
    GridChar(Thin, Invisible, Thick, Thick): "┪",
    GridChar(Thick, Thin, Invisible, Thick): "┹",
    GridChar(Thick, Thick, Invisible, Thin): "┺",
    GridChar(Thin, Thick, Thick, Invisible): "┢",
    GridChar(Invisible, Thin, Thick, Thick): "┱",
    GridChar(Thick, Invisible, Thin, Thick): "┩",
    GridChar(Thick, Thin, Thick, Invisible): "┠",
    GridChar(Invisible, Thick, Thin, Thick): "┯",
    GridChar(Thick, Invisible, Thick, Thin): "┨",
    GridChar(Thin, Thick, Invisible, Thick): "┷",
    GridChar(Thick, Thin, Thin, Thin): "╀",
    GridChar(Thin, Thick, Thin, Thin): "┾",
    GridChar(Thin, Thin, Thick, Thin): "╁",
    GridChar(Thin, Thin, Thin, Thick): "┽",
    GridChar(Thick, Thick, Thin, Thin): "╄",
    GridChar(Thin, Thick, Thick, Thin): "╆",
    GridChar(Thin, Thin, Thick, Thick): "╅",
    GridChar(Thick, Thin, Thin, Thick): "╃",
    GridChar(Thin, Thick, Thin, Thick): "┿",
    GridChar(Thick, Thin, Thick, Thin): "╂",
    GridChar(Thick, Thick, Thick, Thin): "╊",
    GridChar(Thin, Thick, Thick, Thick): "╈",
    GridChar(Thick, Thin, Thick, Thick): "╉",
    GridChar(Thick, Thick, Thin, Thick): "╇",
    # HalfBlockUpperRight
    GridChar(HalfBlockUpperRight, Invisible, HalfBlockUpperRight, Invisible): "▐",
    GridChar(Invisible, HalfBlockUpperRight, Invisible, HalfBlockUpperRight): "▀",
    GridChar(HalfBlockUpperRight, HalfBlockUpperRight, Invisible, Invisible): "▝",
    GridChar(Invisible, HalfBlockUpperRight, HalfBlockUpperRight, Invisible): "▐",
    GridChar(Invisible, Invisible, HalfBlockUpperRight, HalfBlockUpperRight): "▜",
    GridChar(HalfBlockUpperRight, Invisible, Invisible, HalfBlockUpperRight): "▀",
    GridChar(
        HalfBlockUpperRight, HalfBlockUpperRight, HalfBlockUpperRight, Invisible
    ): "▐",
    GridChar(
        Invisible, HalfBlockUpperRight, HalfBlockUpperRight, HalfBlockUpperRight
    ): "▜",
    GridChar(
        HalfBlockUpperRight, Invisible, HalfBlockUpperRight, HalfBlockUpperRight
    ): "▜",
    GridChar(
        HalfBlockUpperRight, HalfBlockUpperRight, Invisible, HalfBlockUpperRight
    ): "▀",
    GridChar(
        HalfBlockUpperRight,
        HalfBlockUpperRight,
        HalfBlockUpperRight,
        HalfBlockUpperRight,
    ): "▜",
    # HalfBlockLowerLeft
    GridChar(HalfBlockLowerLeft, Invisible, HalfBlockLowerLeft, Invisible): "▌",
    GridChar(Invisible, HalfBlockLowerLeft, Invisible, HalfBlockLowerLeft): "▄",
    GridChar(HalfBlockLowerLeft, HalfBlockLowerLeft, Invisible, Invisible): "▙",
    GridChar(Invisible, HalfBlockLowerLeft, HalfBlockLowerLeft, Invisible): "▄",
    GridChar(Invisible, Invisible, HalfBlockLowerLeft, HalfBlockLowerLeft): "▖",
    GridChar(HalfBlockLowerLeft, Invisible, Invisible, HalfBlockLowerLeft): "▌",
    GridChar(
        HalfBlockLowerLeft, HalfBlockLowerLeft, HalfBlockLowerLeft, Invisible
    ): "▙",
    GridChar(
        Invisible, HalfBlockLowerLeft, HalfBlockLowerLeft, HalfBlockLowerLeft
    ): "▄",
    GridChar(
        HalfBlockLowerLeft, Invisible, HalfBlockLowerLeft, HalfBlockLowerLeft
    ): "▌",
    GridChar(
        HalfBlockLowerLeft, HalfBlockLowerLeft, Invisible, HalfBlockLowerLeft
    ): "▙",
    GridChar(
        HalfBlockLowerLeft, HalfBlockLowerLeft, HalfBlockLowerLeft, HalfBlockLowerLeft
    ): "▙",
    # HalfBlock Combos
    GridChar(Invisible, HalfBlockUpperRight, HalfBlockLowerLeft, Invisible): "▛",
    GridChar(HalfBlockUpperRight, Invisible, Invisible, HalfBlockLowerLeft): "▟",
    GridChar(HalfBlockLowerLeft, Invisible, Invisible, HalfBlockUpperRight): "▘",
    GridChar(Invisible, HalfBlockLowerLeft, HalfBlockUpperRight, Invisible): "▗",
    # Halfblock/Thin combos
    GridChar(HalfBlockLowerLeft, Thin, HalfBlockLowerLeft, Invisible): "▌",
    GridChar(Invisible, HalfBlockLowerLeft, Thin, HalfBlockLowerLeft): "▄",
    GridChar(HalfBlockLowerLeft, Invisible, HalfBlockLowerLeft, Thin): "▌",
    GridChar(Thin, HalfBlockLowerLeft, Invisible, HalfBlockLowerLeft): "▄",
    GridChar(HalfBlockUpperRight, Thin, HalfBlockUpperRight, Invisible): "▐",
    GridChar(Invisible, HalfBlockUpperRight, Thin, HalfBlockUpperRight): "▀",
    GridChar(HalfBlockUpperRight, Invisible, HalfBlockUpperRight, Thin): "▐",
    GridChar(Thin, HalfBlockUpperRight, Invisible, HalfBlockUpperRight): "▀",
    # EighthBlockUpperRight
    GridChar(EighthBlockUpperRight, Invisible, EighthBlockUpperRight, Invisible): "▕",
    GridChar(Invisible, EighthBlockUpperRight, Invisible, EighthBlockUpperRight): "▔",
    GridChar(EighthBlockUpperRight, EighthBlockUpperRight, Invisible, Invisible): " ",
    GridChar(Invisible, EighthBlockUpperRight, EighthBlockUpperRight, Invisible): "▕",
    GridChar(Invisible, Invisible, EighthBlockUpperRight, EighthBlockUpperRight): "▕",
    GridChar(EighthBlockUpperRight, Invisible, Invisible, EighthBlockUpperRight): "▔",
    GridChar(
        EighthBlockUpperRight, EighthBlockUpperRight, EighthBlockUpperRight, Invisible
    ): "▕",
    GridChar(
        Invisible, EighthBlockUpperRight, EighthBlockUpperRight, EighthBlockUpperRight
    ): "▔",
    GridChar(
        EighthBlockUpperRight, Invisible, EighthBlockUpperRight, EighthBlockUpperRight
    ): "▕",
    GridChar(
        EighthBlockUpperRight, EighthBlockUpperRight, Invisible, EighthBlockUpperRight
    ): "▔",
    GridChar(
        EighthBlockUpperRight,
        EighthBlockUpperRight,
        EighthBlockUpperRight,
        EighthBlockUpperRight,
    ): "▕",
    # EighthBlockLowerLeft
    GridChar(EighthBlockLowerLeft, Invisible, EighthBlockLowerLeft, Invisible): "▏",
    GridChar(Invisible, EighthBlockLowerLeft, Invisible, EighthBlockLowerLeft): "▁",
    GridChar(EighthBlockLowerLeft, EighthBlockLowerLeft, Invisible, Invisible): "▏",
    GridChar(Invisible, EighthBlockLowerLeft, EighthBlockLowerLeft, Invisible): "▁",
    GridChar(Invisible, Invisible, EighthBlockLowerLeft, EighthBlockLowerLeft): " ",
    GridChar(EighthBlockLowerLeft, Invisible, Invisible, EighthBlockLowerLeft): "▏",
    GridChar(
        EighthBlockLowerLeft, EighthBlockLowerLeft, EighthBlockLowerLeft, Invisible
    ): "▏",
    GridChar(
        Invisible, EighthBlockLowerLeft, EighthBlockLowerLeft, EighthBlockLowerLeft
    ): "▁",
    GridChar(
        EighthBlockLowerLeft, Invisible, EighthBlockLowerLeft, EighthBlockLowerLeft
    ): "▏",
    GridChar(
        EighthBlockLowerLeft, EighthBlockLowerLeft, Invisible, EighthBlockLowerLeft
    ): "▁",
    GridChar(
        EighthBlockLowerLeft,
        EighthBlockLowerLeft,
        EighthBlockLowerLeft,
        EighthBlockLowerLeft,
    ): "▏",
    # EighthBlock Combos
    GridChar(Invisible, EighthBlockLowerLeft, EighthBlockUpperRight, Invisible): " ",
    GridChar(EighthBlockLowerLeft, Invisible, Invisible, EighthBlockUpperRight): " ",
    # Eighthblock/Thin combos
    GridChar(EighthBlockLowerLeft, Thin, EighthBlockLowerLeft, Invisible): "▏",
    GridChar(Invisible, EighthBlockLowerLeft, Thin, EighthBlockLowerLeft): "▁",
    GridChar(EighthBlockLowerLeft, Invisible, EighthBlockLowerLeft, Thin): "▏",
    GridChar(Thin, EighthBlockLowerLeft, Invisible, EighthBlockLowerLeft): "▁",
    GridChar(EighthBlockUpperRight, Thin, EighthBlockUpperRight, Invisible): "▕",
    GridChar(Invisible, EighthBlockUpperRight, Thin, EighthBlockUpperRight): "▔",
    GridChar(EighthBlockUpperRight, Invisible, EighthBlockUpperRight, Thin): "▕",
    GridChar(Thin, EighthBlockUpperRight, Invisible, EighthBlockUpperRight): "▔",
    GridChar(Invisible, Invisible, EighthBlockUpperRight, EighthBlockLowerLeft): "▁",
    GridChar(EighthBlockLowerLeft, EighthBlockUpperRight, Invisible, Invisible): "▔",
}


def grid_char(key: "GridChar") -> "str":
    """Return the character represented by a combination of :class:`LineStyles`."""
    if key in _GRID_CHARS:
        return _GRID_CHARS[key]
    else:
        # If there is not matching character representation, replacing each line style
        # with its parent style in turn and using the result until a matching character
        # is found
        m_key = list(key)
        while any(x.parent for x in m_key):
            for part in sorted(
                {x for x in m_key if x.parent}, key=lambda x: x.rank, reverse=True
            ):
                m_key = [
                    (part.parent or x) if x is part and x.parent else x for x in m_key
                ]
                char = _GRID_CHARS.get(GridChar(*m_key))
                if char:
                    return char
        # If all else fails, return a space
        else:
            return " "


class GridStyle:
    """A collection of characters which can be used to draw a grid."""

    class _BorderLineChars(NamedTuple):
        LEFT: "str"
        MID: "str"
        SPLIT: "str"
        RIGHT: "str"

    def __init__(self, line_style: "LineStyle" = Invisible, mask: "Mask" = Masks.grid):
        """Creates a new :py:class:`GridStyle` instance.

        Args:
            line_style: The line style to use to construct the grid
            mask: A mask which can be used to exclude certain character from the grid
        """
        self.grid = {
            part: GridChar(*((line_style if x else Invisible) for x in mask.mask[part]))
            for part in GridPart
        }

    @property
    def TOP(self) -> "_BorderLineChars":
        """Allow dotted attribute access to the top grid row."""
        return self._BorderLineChars(
            grid_char(self.grid[GridPart.TOP_LEFT]),
            grid_char(self.grid[GridPart.TOP_MID]),
            grid_char(self.grid[GridPart.TOP_SPLIT]),
            grid_char(self.grid[GridPart.TOP_RIGHT]),
        )

    @property
    def MID(self) -> "_BorderLineChars":
        """Allow dotted attribute access to the mid grid row."""
        return self._BorderLineChars(
            grid_char(self.grid[GridPart.MID_LEFT]),
            grid_char(self.grid[GridPart.MID_MID]),
            grid_char(self.grid[GridPart.MID_SPLIT]),
            grid_char(self.grid[GridPart.MID_RIGHT]),
        )

    @property
    def SPLIT(self) -> "_BorderLineChars":
        """Allow dotted attribute access to the split grid row."""
        return self._BorderLineChars(
            grid_char(self.grid[GridPart.SPLIT_LEFT]),
            grid_char(self.grid[GridPart.SPLIT_MID]),
            grid_char(self.grid[GridPart.SPLIT_SPLIT]),
            grid_char(self.grid[GridPart.SPLIT_RIGHT]),
        )

    @property
    def BOTTOM(self) -> "_BorderLineChars":
        """Allow dotted attribute access to the bottom grid row."""
        return self._BorderLineChars(
            grid_char(self.grid[GridPart.BOTTOM_LEFT]),
            grid_char(self.grid[GridPart.BOTTOM_MID]),
            grid_char(self.grid[GridPart.BOTTOM_SPLIT]),
            grid_char(self.grid[GridPart.BOTTOM_RIGHT]),
        )

    @property
    def HORIZONTAL(self) -> "str":
        """For compatibility with :class:`prompt_toolkit.widgets.base.Border`."""
        return self.SPLIT_MID

    @property
    def VERTICAL(self) -> "str":
        """For compatibility with :class:`prompt_toolkit.widgets.base.Border`."""
        return self.MID_SPLIT

    def __getattr__(self, value: "str") -> "str":
        """Allows the parts of the grid to be accessed as attributes by name."""
        key = getattr(GridPart, value)
        return grid_char(self.grid[key])

    def __dir__(self) -> "list":
        """List the public attributes of the grid style."""
        return [x.name for x in GridPart] + ["TOP", "MID", "SPLIT", "BOTTOM"]

    def __add__(self, other: "GridStyle") -> "GridStyle":
        """Combine two grid styles."""
        grid_style = GridStyle()
        grid_style.grid = {}
        for part in GridPart:
            grid_style.grid[part] = GridChar(
                *(max(self.grid[part][i], other.grid[part][i]) for i in range(4))
            )
        return grid_style

    def __repr__(self) -> "str":
        """Returns a string representation of the grid style."""
        return "\n".join(
            "".join(
                grid_char(char_key)
                for char_key in list(self.grid.values())[i * 4 : (i + 1) * 4]
            )
            for i in range(4)
        )


InnerEdgeGridStyle = (
    EighthBlockLowerLeft.top_edge
    + EighthBlockLowerLeft.right_edge
    + EighthBlockUpperRight.left_edge
    + EighthBlockUpperRight.bottom_edge
    + Thin.inner
)


OuterEdgeGridStyle = (
    EighthBlockLowerLeft.top_edge
    + EighthBlockUpperRight.right_edge
    + EighthBlockUpperRight.bottom_edge
    + EighthBlockLowerLeft.left_edge
    + Thin.inner
)

HalfBlockOuterGridStyle = (
    HalfBlockUpperRight.top_edge
    + HalfBlockUpperRight.right_edge
    + HalfBlockLowerLeft.left_edge
    + HalfBlockLowerLeft.bottom_edge
    + Thin.inner
)

HalfBlockInnerGridStyle = (
    HalfBlockLowerLeft.top_edge
    + HalfBlockLowerLeft.right_edge
    + HalfBlockUpperRight.left_edge
    + HalfBlockUpperRight.bottom_edge
    + Thin.inner
)


class WeightedLineStyle(NamedTuple):
    """A :class:`LineStyle` with a weight."""

    weight: "int"
    value: "LineStyle"


class WeightedInt(NamedTuple):
    """An :class:`int` with a weight."""

    weight: "int"
    value: "int"


class BorderLineStyle(NamedTuple):
    """A description of a cell border: a :class:`LineStyle` for each edge."""

    top: "Optional[LineStyle]" = None
    right: "Optional[LineStyle]" = None
    bottom: "Optional[LineStyle]" = None
    left: "Optional[LineStyle]" = None


class WeightedBorderLineStyle(NamedTuple):
    """A weighted description of a cell border: weighted values for each edge."""

    top: "WeightedLineStyle"
    right: "WeightedLineStyle"
    bottom: "WeightedLineStyle"
    left: "WeightedLineStyle"

    # We cannot use :py:func:`functools.cached_property` here as it does not work with
    # :py:Class:`NamedTuple`s.
    @property  # type: ignore
    @lru_cache(maxsize=1)  # noqa: B019
    def border_line_style(self) -> "BorderLineStyle":
        """Get the unweighted border line style."""
        return BorderLineStyle(*(x.value for x in self))


class Padding(NamedTuple):
    """A weighted description of a cell padding: weighted values for each edge."""

    top: "Optional[int]"
    right: "Optional[int]"
    bottom: "Optional[int]"
    left: "Optional[int]"


class WeightedPadding(NamedTuple):
    """A description of a cell padding: :class:`LineStyle`s for each edge."""

    top: "WeightedInt"
    right: "WeightedInt"
    bottom: "WeightedInt"
    left: "WeightedInt"

    # We cannot use :py:func:`functools.cached_property` here as it does not work with
    # :py:Class:`NamedTuple`s.
    @property  # type: ignore
    @lru_cache(maxsize=1)  # noqa: B019
    def padding(self) -> "Padding":
        """Get the padding without weights."""
        return Padding(*(x.value for x in self))


class BorderVisibility(NamedTuple):
    """Defines if each border edge is visible."""

    top: "bool"
    right: "bool"
    bottom: "bool"
    left: "bool"



"""Defines a configuration class for euporie.core."""

from __future__ import annotations

import argparse
import json
import logging
import os
from ast import literal_eval
from collections import ChainMap
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Protocol, TextIO, cast

import fastjsonschema
from appdirs import user_config_dir
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.filters.utils import to_filter
from prompt_toolkit.utils import Event
from upath import UPath

from euporie.core import __app_name__, __copyright__, __version__
from euporie.core.commands import add_cmd, get_cmd

if TYPE_CHECKING:
    from typing import IO, Any, Callable, Optional, Sequence, Type, Union

    from prompt_toolkit.filters.base import Filter, FilterOrBool

    from euporie.core.widgets.menu import MenuItem


class ConfigurableApp(Protocol):
    """An application with configuration."""

    config: "Config"


log = logging.getLogger(__name__)


class ArgumentParser(argparse.ArgumentParser):
    """An argument parser which lexes and formats the help message before printing it."""

    def _print_message(
        self, message: "str", file: "Optional[IO[str]]" = None
    ) -> "None":
        from prompt_toolkit.formatted_text.base import FormattedText
        from prompt_toolkit.lexers.pygments import _token_cache
        from prompt_toolkit.shortcuts.utils import print_formatted_text
        from prompt_toolkit.styles.pygments import style_from_pygments_cls

        from euporie.core.pygments import ArgparseLexer, EuporiePygmentsStyle

        if message:
            file = cast("Optional[TextIO]", file)
            print_formatted_text(
                FormattedText(
                    [
                        (_token_cache[t], v)
                        for _, t, v in ArgparseLexer().get_tokens_unprocessed(message)
                    ]
                ),
                file=file,
                style=style_from_pygments_cls(EuporiePygmentsStyle),
                include_default_pygments_style=False,
            )


class BooleanOptionalAction(argparse.Action):
    """Action for boolean flags.

    Included because `argparse.BooleanOptionalAction` is not present in `python<=3.9`.
    """

    def __init__(self, option_strings: "list[str]", *args: "Any", **kwargs: "Any"):
        """Initiate the Action, as per `argparse.BooleanOptionalAction`."""
        _option_strings = list(option_strings)
        for option_string in option_strings:
            if option_string.startswith("--"):
                _option_strings.append(f"--no-{option_string[2:]}")
        kwargs["nargs"] = 0
        super().__init__(_option_strings, *args, **kwargs)

    def format_usage(self) -> "str":
        """Formats the action string.

        Returns:
            The formatted string.

        """
        return " | ".join(self.option_strings)

    def __call__(
        self,
        parser: "argparse.ArgumentParser",
        namespace: "argparse.Namespace",
        values: "Union[str, Sequence[Any], None]",
        option_string: "Optional[str]" = None,
    ) -> "None":
        """Set the value to True or False depending on the flag provided."""
        if option_string in self.option_strings:
            assert isinstance(option_string, str)
            setattr(namespace, self.dest, not option_string.startswith("--no-"))


TYPE_ACTIONS: "dict[Callable[[Any], Any], Type[argparse.Action]]" = {
    bool: BooleanOptionalAction
}


class JSONEncoderPlus(json.JSONEncoder):
    """JSON encode class which encodes paths as strings."""

    def default(self, o: "Any") -> "Union[bool, int, float, str, None]":
        """Encode an object to JSON.

        Args:
            o: The object to encode

        Returns:
            The encoded object

        """
        if isinstance(o, (Path, UPath)):
            return str(o)
        return json.JSONEncoder.default(self, o)


_json_encoder = JSONEncoderPlus()


class Config:
    """A configuration store."""

    settings: "dict[str, Setting]" = {}
    conf_file_name = "config.json"

    def __init__(self) -> "None":
        """Create a new configuration object instance."""
        self.app_name: "str" = "base"
        self.app_cls: "Optional[Type[ConfigurableApp]]" = None

    def _save(self, setting: "Setting") -> "None":
        """Save settings to user's configuration file."""
        json_data = self.load_config_file()
        json_data.setdefault(self.app_name, {})[setting.name] = setting.value
        if self.valid_user:
            log.debug("Saving setting `%s`", setting)
            with open(self.config_file_path, "w") as f:
                json.dump(json_data, f, indent=2)

    def load(self, cls: "Type[ConfigurableApp]", app_name: "str" = "") -> "None":
        """Loads the command line, environment, and user configuration."""
        self.app_name = app_name or cls.__module__.split(".")[-1]
        self.app_cls = cls
        log.debug("Loading config for %s", self.app_name)

        user_conf_dir = Path(user_config_dir(__app_name__, appauthor=False))
        user_conf_dir.mkdir(exist_ok=True, parents=True)
        self.config_file_path = (user_conf_dir / self.conf_file_name).with_suffix(
            ".json"
        )
        self.valid_user = True

        set_values = ChainMap(
            self.load_args(),
            # Load app specific env vars
            self.load_env(app_name=self.app_name),
            # Load global env vars
            self.load_env(),
            # Load app specific user config
            self.load_user(app_name=self.app_name),
            # Load global user config
            self.load_user(),
        )
        for name, setting in Config.settings.items():
            if setting.name in set_values:
                # Set value without triggering hooks
                setting._value = set_values[name]
            setting.event += self._save

    @property
    def schema(self) -> "dict[str, Any]":
        """Return a JSON schema for the config."""
        return {
            "title": "Euporie Configuration",
            "description": "A configuration for euporie",
            "type": "object",
            "properties": {name: item.schema for name, item in self.settings.items()},
        }

    def load_parser(self) -> "argparse.ArgumentParser":
        """Constructs an :py:class:`ArgumentParser`."""
        parser = ArgumentParser(
            description=self.app_cls.__doc__,
            epilog=__copyright__,
            allow_abbrev=True,
            formatter_class=argparse.MetavarTypeHelpFormatter,
        )
        # Add options to the relevant subparsers
        for setting in self.settings.values():
            args, kwargs = setting.parser_args
            parser.add_argument(*args, **kwargs)
        return parser

    def load_args(self) -> "dict[str, Any]":
        """Attempts to load configuration settings from commandline flags."""
        result = {}
        namespace, _ = self.load_parser().parse_known_intermixed_args()
        for name, value in vars(namespace).items():
            if value is not None:
                # Convert to json and back to attain json types
                json_data = json.loads(_json_encoder.encode({name: value}))
                try:
                    fastjsonschema.validate(self.schema, json_data)
                except fastjsonschema.JsonSchemaValueException as error:
                    log.warning(f"Error in command line parameter `{name}`: {error}")
                else:
                    result[name] = value
        return result

    def load_env(self, app_name: "str" = "") -> "dict[str, Any]":
        """Attempt to load configuration settings from environment variables."""
        result = {}
        for name, setting in self.settings.items():
            if app_name:
                env = f"{__app_name__.upper()}_{self.app_name.upper()}_{setting.name.upper()}"
            else:
                env = f"{__app_name__.upper()}_{setting.name.upper()}"
            if env in os.environ:
                value = os.environ.get(env)
                parsed_value: "Any" = value
                # Attempt to parse the value as a literal
                if value:
                    try:
                        parsed_value = literal_eval(value)
                    except (
                        ValueError,
                        TypeError,
                        SyntaxError,
                        MemoryError,
                        RecursionError,
                    ):
                        pass
                # Attempt to cast the value to the desired type
                try:
                    parsed_value = setting.type(value)
                except (ValueError, TypeError):
                    log.warning(
                        f"Environment variable `{env}` not understood"
                        f" - `{setting.type.__name__}` expected"
                    )
                else:
                    json_data = json.loads(_json_encoder.encode({name: parsed_value}))
                    try:
                        fastjsonschema.validate(self.schema, json_data)
                    except fastjsonschema.JsonSchemaValueException as error:
                        log.error(f"Error in environment variable: `{env}`\n{error}")
                    else:
                        result[name] = parsed_value
        return result

    def load_config_file(self) -> "dict[str, Any]":
        """Attempt to load JSON configuration file."""
        results = {}
        assert isinstance(self.config_file_path, Path)
        if self.valid_user and self.config_file_path.exists():
            with open(self.config_file_path, "r") as f:
                try:
                    json_data = json.load(f)
                except json.decoder.JSONDecodeError:
                    log.error(
                        "Could not parse the configuration file: "
                        f"{self.config_file_path}\n"
                        "Is it valid json?"
                    )
                    self.valid_user = False
                else:
                    results.update(json_data)
        return results

    def load_user(self, app_name: "str" = "") -> "dict[str, Any]":
        """Attempt to load JSON configuration file."""
        results = {}
        # Load config file
        json_data = self.load_config_file()
        # Load section for the current app
        if app_name:
            json_data = json_data.get(self.app_name, {})
        # Validate the configuration file
        try:
            # Validate a copy so the original data is not modified
            fastjsonschema.validate(self.schema, dict(json_data))
        except fastjsonschema.JsonSchemaValueException as error:
            log.warning(f"Error in config file: `{self.config_file_path}`: {error}")
            self.valid_user = False
        else:
            results.update(json_data)
        return results

    def get(self, name: "str") -> "Any":
        """Access a configuration value, falling back to the default value if unset.

        Args:
            name: The name of the attribute to access.

        Returns:
            The configuration variable value.

        """
        return self.settings[name].value

    def get_item(self, name: "str") -> "Any":
        """Access a configuration item.

        Args:
            name: The name of the attribute to access.

        Returns:
            The configuration item.

        """
        return self.settings.get(name)

    def filter(self, name: "str") -> "Filter":
        """Return a :py:class:`Filter` for a configuration item."""
        return Condition(partial(self.get, name))

    def __getattr__(self, name: "str") -> "Any":
        """Enables access of config elements via dotted attributes.

        Args:
            name: The name of the attribute to access.

        Returns:
            The configuration variable value.

        """
        return self.get(name)

    def __setitem__(self, name: "str", value: "Any") -> "None":
        """Set a configuration attribute.

        Args:
            name: The name of the attribute to set.
            value: The value to give the attribute.

        """
        setting = self.settings[name]
        setting.value = value


class Setting:
    """A single configuration item."""

    def __init__(
        self,
        name: "str",
        default: "Any",
        help_: "str",
        description: "str",
        type_: "Optional[Callable[[Any], Any]]" = None,
        title: "Optional[str]" = None,
        choices: "Optional[list[Any]]" = None,
        action: "Optional[Union[argparse.Action,str]]" = None,
        flags: "Optional[list[str]]" = None,
        schema: "Optional[dict[str, Any]]" = None,
        nargs: "Optional[str|int]" = None,
        hidden: "FilterOrBool" = False,
        hooks: "Optional[list[Callable[[Setting], None]]]" = None,
        cmd_filter: "FilterOrBool" = True,
        **kwargs: "Any",
    ) -> "None":
        """Create a new configuration item."""
        self.name = name
        self.default = default
        self._value = default
        self.title = title or self.name.replace("_", " ")
        self.help = help_
        self.description = description
        self.choices = choices
        self.type = type_ or type(default)
        self.action = action or TYPE_ACTIONS.get(self.type)
        self.flags = flags or [f"--{name.replace('_','-')}"]
        self._schema: "dict[str, Any]" = {
            "type": {
                bool: "boolean",
                str: "string",
                int: "integer",
                float: "float",
                UPath: "string",
            }.get(self.type),
            **(schema or {}),
        }
        self.nargs = nargs
        self.hidden = to_filter(hidden)
        self.kwargs = kwargs
        self.cmd_filter = cmd_filter

        self.event = Event(self)
        for hook in hooks or []:
            self.event += hook

        self.register_commands()

    def register_commands(self) -> "None":
        """Register commands to set this setting."""
        name = self.name.replace("_", "-")
        if self.type in (bool, int) or self.choices is not None:
            toggled_filter = None
            if self.type == bool:

                def _toggled() -> "bool":
                    from euporie.core.app import get_app

                    app = get_app()
                    value = app.config.get(self.name)
                    return bool(getattr(app.config, self.name, not value))

                toggled_filter = Condition(_toggled)

            add_cmd(
                name=f"toggle-{name}",
                toggled=toggled_filter,
                hidden=self.hidden,
                title=f"Toggle {self.title}",
                menu_title=self.kwargs.get("menu_title"),
                description=f'Toggle the value of the "{self.name}" configuration option.',
                filter=self.cmd_filter,
            )(self.toggle)

        for choice in (self.choices or self.schema.get("enum", [])) or []:
            add_cmd(
                name=f"set-{name}-{choice}",
                hidden=self.hidden,
                toggled=Condition(partial(lambda x: self.value == x, choice)),
                title=f"Set {self.title} to {choice}",
                menu_title=str(choice).replace("_", " ").capitalize(),
                description=f'Set the value of the "{self.name}" '
                f'configuration option to "{choice}"',
                filter=self.cmd_filter,
            )(partial(setattr, self, "value", choice))

    def toggle(self) -> "None":
        """Toggle the setting's value."""
        if self.type == bool:
            new = not self.value
        elif (
            self.type == int
            and "minimum" in (schema := self.schema)
            and "maximum" in schema
        ):
            new = schema["minimum"] + (self.value - schema["minimum"] + 1) % (
                schema["maximum"] + 1
            )
        elif self.choices is not None:
            new = self.choices[(self.choices.index(self.value) + 1) % len(self.choices)]
        else:
            raise NotImplementedError
        self.value = new

    @property
    def value(self) -> "Any":
        """Return the current value."""
        return self._value

    @value.setter
    def value(self, new: "Any") -> "None":
        """Set the current value."""
        self._value = new
        self.event.fire()

    @property
    def schema(self) -> "dict[str, Any]":
        """Return a json schema property for the config item."""
        return {
            "description": self.help,
            **({"enum": self.choices} if self.choices is not None else {}),
            **({"default": self.default} if self.default is not None else {}),
            **self._schema,
        }

    @property
    def menu(self) -> "MenuItem":
        """Return a menu item for the setting."""
        from euporie.core.widgets.menu import MenuItem

        choices = (self.choices or self.schema.get("enum", [])) or []
        if choices:
            return MenuItem(
                self.title.capitalize(),
                children=[
                    cmd.menu
                    for cmd in sorted(
                        (
                            get_cmd(f"set-{self.name.replace('_', '-')}-{choice}")
                            for choice in choices
                        ),
                        key=lambda x: x.menu_title,
                    )
                ],
            )
        elif self.type in (bool, int):
            return get_cmd(f"toggle-{self.name.replace('_', '-')}").menu
        else:
            raise NotImplementedError

    @property
    def parser_args(self) -> "tuple[list[str], dict[str, Any]]":
        """Return arguments for construction of an :class:`argparse.ArgumentParser`."""
        # Do not set defaults for command line arguments, as default values
        # would override values set in the configuration file
        args = self.flags or [self.name]

        kwargs: "dict[str, Any]" = {
            "action": self.action,
            "help": self.help,
        }

        if self.nargs:
            kwargs["nargs"] = self.nargs
        if self.type is not None and self.name != "version":
            kwargs["type"] = self.type
        if self.choices:
            kwargs["choices"] = self.choices
        if "version" in self.kwargs:
            kwargs["version"] = self.kwargs["version"]
        if "const" in self.kwargs:
            kwargs["const"] = self.kwargs["const"]

        return args, kwargs

    def __repr__(self):
        return f"<Setting {self.name}={self.value.__repr__()}>"


def add_setting(
    name: "str",
    default: "Any",
    help_: "str",
    description: "str",
    type_: "Optional[Callable[[Any], Any]]" = None,
    action: "Optional[Union[argparse.Action,str]]" = None,
    flags: "Optional[list[str]]" = None,
    schema: "Optional[dict[str, Any]]" = None,
    nargs: "Optional[str|int]" = None,
    hidden: "FilterOrBool" = False,
    hooks: "Optional[list[Callable[[Setting], None]]]" = None,
    cmd_filter: "FilterOrBool" = True,
    **kwargs: "Any",
) -> "None":
    """Register a new config item."""
    Config.settings[name] = Setting(
        name=name,
        default=default,
        help_=help_,
        description=description,
        type_=type_,
        action=action,
        flags=flags,
        schema=schema,
        nargs=nargs,
        hidden=hidden,
        hooks=hooks,
        cmd_filter=cmd_filter,
        **kwargs,
    )


# ################################### Settings ####################################

add_setting(
    name="version",
    default=False,
    flags=["--version", "-V"],
    action="version",
    hidden=True,
    version=f"%(prog)s {__version__}",
    help_="Show the version number and exit",
    description="""
        If set, euporie will print the current version number of the application and exit.
        All other configuration options will be ignored.

        .. note::

           This cannot be set in the configuration file or via an environment variable
    """,
)



"""Contains functions which convert data to formatted text."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.formatted_text import to_formatted_text

from euporie.core.convert.base import register
from euporie.core.formatted_text.ansi import ANSI

if TYPE_CHECKING:
    from typing import Optional

    from prompt_toolkit.formatted_text.base import StyleAndTextTuples
    from upath import UPath

    from euporie.core.formatted_text.html import HTML

_html_cache: "SimpleCache[int, HTML]" = SimpleCache(maxsize=20)


@register(
    from_="html",
    to="formatted_text",
)
def html_to_ft(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "StyleAndTextTuples":
    """Converts markdown to formatted text."""
    from euporie.core.formatted_text.html import HTML

    html = _html_cache.get(hash(data), partial(HTML, data, width=width, base=path))
    if (
        width is not None
        and height is not None
        and (html.width != width or html.height != height)
    ):
        html.render(width, height)
    return to_formatted_text(HTML(data, width=width, base=path))


@register(
    from_="ansi",
    to="formatted_text",
)
def ansi_to_ft(
    data: "str",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "StyleAndTextTuples":
    """Converts ANSI text to formatted text."""
    return to_formatted_text(ANSI(data.strip()))



"""Contains main format conversion function."""

from __future__ import annotations

import logging
import mimetypes
from functools import partial
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.cache import FastDictCache, SimpleCache
from prompt_toolkit.filters import to_filter
from upath import UPath
from upath.implementations.http import HTTPPath

if TYPE_CHECKING:
    from typing import Any, Callable, Iterable, Optional, Union

    from prompt_toolkit.filters import Filter, FilterOrBool

log = logging.getLogger(__name__)


MIME_FORMATS = {
    "image/svg+xml": "svg",
    "image/png": "png",
    "image/jpeg": "jpeg",
    "image/gif": "gif",
    "application/pdf": "pdf",
    "text/html": "html",
    "text/latex": "latex",
    "application/x-latex": "latex",
    "text/markdown": "markdown",
    "text/x-markdown": "markdown",
    "text/*": "ansi",
    "stream/std*": "ansi",
    "*": "ansi",
}

BASE64_FORMATS = {"png", "jepg", "pdf", "gif"}


def get_format(path: "UPath|str", default: "str" = "") -> "str":
    """Attempt to guess the format of a path."""
    if isinstance(path, str):
        path = UPath(path)
    if not default:
        if isinstance(path, HTTPPath):
            default = "html"
        else:
            default = "ansi"
    mime, _ = mimetypes.guess_type(path)
    return MIME_FORMATS.get(mime, default)


class Converter(NamedTuple):
    """Holds a conversion function and its weight."""

    func: Callable
    filter_: Filter
    weight: int = 1


converters: "dict[str, dict[str, list[Converter]]]" = {}

_CONVERSION_CACHE: "SimpleCache" = SimpleCache(maxsize=2048)
_FILTER_CACHE: "SimpleCache" = SimpleCache()


def register(
    from_: "Union[Iterable[str], str]",
    to: "str",
    filter_: "FilterOrBool" = True,
    weight: "int" = 1,
) -> "Callable":
    """Adds a converter to the centralized format conversion system."""
    if isinstance(from_, str):
        from_ = (from_,)

    def decorator(func: "Callable") -> "Callable":
        if to not in converters:
            converters[to] = {}
        for from_format in from_:
            if from_format not in converters[to]:
                converters[to][from_format] = []
            converters[to][from_format].append(
                Converter(func=func, filter_=to_filter(filter_), weight=weight)
            )
        return func

    return decorator


def find_route(from_: "str", to: "str") -> "Optional[list]":
    """Finds the shortest conversion path between two formats."""
    if from_ == to:
        return [from_]

    chains = []

    def find(start: "str", chain: "list[str]") -> "None":
        if chain[0] == start:
            chains.append(chain)
        sources: "dict[str, list[Converter]]" = converters.get(chain[0], {})
        for link in sources:
            if link not in chain:
                if any(
                    _FILTER_CACHE.get((conv,), conv.filter_)
                    for conv in sources.get(link, [])
                ):
                    find(start, [link, *chain])

    find(from_, [to])

    if chains:
        # Find chain with shortest weighted length
        return sorted(
            chains,
            key=lambda chain: sum(
                [
                    min(
                        [
                            conv.weight
                            for conv in converters.get(step_b, {}).get(step_a, [])
                            if _FILTER_CACHE.get((conv,), conv.filter_)
                        ]
                    )
                    for step_a, step_b in zip(chain, chain[1:])
                ]
            ),
        )[0]
    else:
        return None


_CONVERTOR_ROUTE_CACHE: "FastDictCache[tuple[str, str], Optional[list]]" = (
    FastDictCache(find_route)
)


def convert(
    data: "Any",
    from_: "str",
    to: "str",
    cols: "Optional[int]" = None,
    rows: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "Any":
    """Convert between formats."""
    try:
        data_hash = hash(data)
    except TypeError as error:
        log.exception("Cannot hash %s", data)
        raise error

    def _convert(
        data: "str",
        from_: "str",
        to: "str",
        cols: "Optional[int]" = None,
        rows: "Optional[int]" = None,
        fg: "Optional[str]" = None,
        bg: "Optional[str]" = None,
        path: "Optional[UPath]" = None,
    ) -> "Any":
        if from_ == to:
            return data
        route = _CONVERTOR_ROUTE_CACHE[(from_, to)]
        # log.debug("Converting from '%s' to '%s' using route: %s", from_, to, route)
        if route is None:
            raise NotImplementedError(f"Cannot convert from `{from_}` to `{to}`")
        output = data
        for stage_a, stage_b in zip(route, route[1:]):
            # Find converter with lowest weight
            func = sorted(
                [
                    conv
                    for conv in converters[stage_b][stage_a]
                    if _FILTER_CACHE.get((conv,), conv.filter_)
                ],
                key=lambda x: x.weight,
            )[0].func
            # Add intermediate steps to the cache
            try:
                output_hash = hash(data)
            except TypeError as error:
                log.exception("Cannot hash %s", data)
                raise error
            output = _CONVERSION_CACHE.get(
                (output_hash, from_, stage_b, cols, rows, fg, bg, path),
                partial(func, output, cols, rows, fg, bg, path),
            )
            if output is None:
                log.error(
                    "Failed to convert `%s` from `%s`"
                    " to `%s` using route `%s` at stage `%s`",
                    data.__repr__()[:10],
                    from_,
                    to,
                    route,
                    stage_b,
                )
                return [("", "(Conversion failed")]
        return output

    data = _CONVERSION_CACHE.get(
        (data_hash, from_, to, cols, rows, fg, bg, path),
        partial(_convert, data, from_, to, cols, rows, fg, bg, path),
    )

    return data



"""Buffer processors."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.layout.processors import (
    AppendAutoSuggestion,
    Processor,
    Transformation,
)
from prompt_toolkit.layout.utils import explode_text_fragments

if TYPE_CHECKING:
    from prompt_toolkit.layout.processors import TransformationInput


log = logging.getLogger(__name__)


class AppendLineAutoSuggestion(AppendAutoSuggestion):
    """Append the auto suggestion to the current line of the input."""

    def apply_transformation(self, ti: "TransformationInput") -> "Transformation":
        """Insert fragments at the end of the current line."""
        if ti.lineno == ti.document.cursor_position_row:
            buffer = ti.buffer_control.buffer

            if buffer.suggestion and ti.document.is_cursor_at_the_end_of_line:
                suggestion = buffer.suggestion.text
            else:
                suggestion = ""
            return Transformation(fragments=ti.fragments + [(self.style, suggestion)])
        else:
            return Transformation(fragments=ti.fragments)


class ShowTrailingWhiteSpaceProcessor(Processor):
    """Make trailing whitespace visible."""

    def __init__(
        self,
        char: "str" = "·",
        style: "str" = "class:trailing-whitespace",
    ) -> "None":
        """Create a new processor instance."""
        self.char = char
        self.style = style

    def apply_transformation(self, ti: "TransformationInput") -> "Transformation":
        """Walk backwards through all the fragments and replace whitespace."""
        fragments = ti.fragments
        if fragments and fragments[-1][1].endswith(" "):
            fragments = explode_text_fragments(fragments)
            new_char = self.char
            for i in range(len(fragments) - 1, -1, -1):
                style, char, *_ = fragments[i]
                if char == " ":
                    fragments[i] = (f"{style} {self.style}", new_char)
                else:
                    break
        return Transformation(fragments)



"""Contains functions which convert data to base64 format."""

from __future__ import annotations

import base64
from typing import TYPE_CHECKING

from euporie.core.convert.base import register

if TYPE_CHECKING:
    from typing import Optional, Union

    from upath import UPath


@register(
    from_="png",
    to="base64-png",
    filter_=True,
)
@register(
    from_="jpeg",
    to="base64-jpeg",
    filter_=True,
)
@register(
    from_="pdf",
    to="base64-pdf",
    filter_=True,
)
@register(
    from_="svg",
    to="base64-svg",
    filter_=True,
)
def bytes_to_base64_py(
    data: "Union[str, bytes]",
    width: "Optional[int]" = None,
    height: "Optional[int]" = None,
    fg: "Optional[str]" = None,
    bg: "Optional[str]" = None,
    path: "Optional[UPath]" = None,
) -> "str":
    """Converts bytes to base64 encoded data."""
    if isinstance(data, str):
        data = data.encode()
    return base64.b64encode(data).decode().replace("\n", "").strip()
